import tensorflow as tf
from absl.testing import parameterized

from keras_aug import layers


class SolarizationTest(tf.test.TestCase, parameterized.TestCase):
    @parameterized.named_parameters(
        ("0_255", 0, 255),
        ("64_191", 64, 191),
        ("127_128", 127, 128),
        ("191_64", 191, 64),
        ("255_0", 255, 0),
    )
    def test_output_values(self, input_value, expected_value):
        solarization = layers.RandomSolarize(
            value_range=(0, 255), threshold_factor=(0, 0)
        )

        self._test_input_output(
            layer=solarization,
            input_value=input_value,
            expected_value=expected_value,
            dtype=tf.uint8,
        )

    @parameterized.named_parameters(
        ("0_245", 0, 245),
        ("255_0", 255, 0),
    )
    def test_solarization_with_addition(self, input_value, output_value):
        solarization = layers.RandomSolarize(
            value_range=(0, 255),
            threshold_factor=(0, 0),
            addition_factor=(10.0, 10.0),
        )
        self._test_input_output(
            layer=solarization,
            input_value=input_value,
            expected_value=output_value,
            dtype=tf.float32,
        )

    @parameterized.named_parameters(
        ("0_0", 0, 0),
        ("64_64", 64, 64),
        ("127_127", 127, 127),
        ("191_64", 191, 64),
        ("255_0", 255, 0),
    )
    def test_only_values_above_threshold_are_solarized(
        self, input_value, output_value
    ):
        solarization = layers.RandomSolarize(
            threshold_factor=(128, 128), value_range=(0, 255)
        )

        self._test_input_output(
            layer=solarization,
            input_value=input_value,
            expected_value=output_value,
            dtype=tf.uint8,
        )

    def _test_input_output(self, layer, input_value, expected_value, dtype):
        input = tf.ones(shape=(2, 4, 4, 3), dtype=dtype) * input_value
        expected_output = tf.clip_by_value(
            (
                tf.ones(shape=(2, 4, 4, 3), dtype=layer.compute_dtype)
                * expected_value
            ),
            0,
            255,
        )

        output = layer(input)

        self.assertAllClose(output, expected_output)



import tensorflow as tf
from tensorflow import keras

from keras_aug import layers
from keras_aug.datapoints import image as image_utils
from keras_aug.layers.base.vectorized_base_random_layer import (
    VectorizedBaseRandomLayer,
)
from keras_aug.utils import augmentation as augmentation_utils
from keras_aug.utils.augmentation import H_AXIS
from keras_aug.utils.augmentation import IMAGES
from keras_aug.utils.augmentation import W_AXIS
from keras_aug.utils.distribution import stateless_random_beta
from keras_aug.utils.distribution import stateless_random_dirichlet


@keras.utils.register_keras_serializable(package="keras_aug")
class AugMix(VectorizedBaseRandomLayer):
    """Performs the AugMix data augmentation technique.

    AugMix aims to produce images with variety while preserving the image
    semantics and local statistics. During the augmentation process, each image
    is augmented ``num_chains`` different ways, each way consisting of
    ``chain_depth`` augmentations. Augmentations are sampled from the list:
    [translation, shearing, rotation, posterization, histogram equalization,
    solarization and auto contrast]. The results of each chain are then mixed
    together with the original image based on random samples from a Dirichlet
    distribution.

    Args:
        value_range (Sequence[float]): The range of values the incoming
            images will have. This is typically either ``[0, 1]`` or
            ``[0, 255]`` depending on how your preprocessing pipeline is set up.
        severity (float|(float, float)|keras_aug.FactorSampler, optional): The range
            of the strength of augmentations. When represented as a single float,
            the factor will be picked between ``[0.01, upper]``. Defaults to
            ``[0.01, 0.3]``.
        num_chains (int, optional): The number of different chains to be mixed.
            Defaults to ``3``.
        chain_depth (int, Sequence[int], optional): The range of the number of
            transformations in the chains. Defaults to ``[1, 3]``.
        alpha (float, optional): The probability coefficients for the Beta and
            Dirichlet distributions. Defaults to ``1.0``.
        seed (int|float, optional): The random seed. Defaults to ``None``.

    References:
        - `AugMix <https://arxiv.org/abs/1912.02781>`_
        - `AugMix Official Repo <https://github.com/google-research/augmix>`_
        - `KerasCV <https://github.com/keras-team/keras-cv>`_
    """  # noqa: E501

    def __init__(
        self,
        value_range,
        severity=[0.01, 0.3],
        num_chains=3,
        chain_depth=[1, 3],
        alpha=1.0,
        seed=None,
        **kwargs,
    ):
        super().__init__(seed=seed, **kwargs)
        self.value_range = value_range
        if isinstance(severity, (int, float)):
            severity = (0.01, severity)
        self.severity = augmentation_utils.parse_factor(
            severity,
            min_value=0.01,
            max_value=1.0,
            seed=seed,
        )
        self.num_chains = num_chains
        if isinstance(chain_depth, int):
            chain_depth = [chain_depth, chain_depth]
        self.chain_depth = chain_depth
        self.alpha = alpha
        self.seed = seed

        # initialize layers
        self.auto_contrast = layers.AutoContrast(
            value_range=self.value_range, dtype=self.compute_dtype
        )
        self.equalize = layers.Equalize(
            value_range=self.value_range, dtype=self.compute_dtype
        )

    def get_random_transformation_batch(self, batch_size, **kwargs):
        # cast to float32 to avoid numerical issue
        # sample from dirichlet
        alpha = tf.ones([self.num_chains], dtype=tf.float32) * self.alpha
        chain_mixing_weights = stateless_random_dirichlet(
            (batch_size, self.num_chains),
            seed=self._random_generator.make_seed_for_stateless_op(),
            alpha=alpha,
            dtype=tf.float32,
        )
        # sample from beta
        weight_sample = stateless_random_beta(
            (batch_size, 1),
            seed_alpha=self._random_generator.make_seed_for_stateless_op(),
            seed_beta=self._random_generator.make_seed_for_stateless_op(),
            alpha=self.alpha,
            beta=self.alpha,
            dtype=tf.float32,
        )
        return {
            "chain_mixing_weights": chain_mixing_weights,
            "weight_sample": weight_sample,
        }

    def augment_ragged_image(self, image, transformation, **kwargs):
        images = tf.expand_dims(image, axis=0)
        transformations = augmentation_utils.expand_dict_dims(
            transformation, axis=0
        )
        images = self.augment_images(
            images=images, transformations=transformations, **kwargs
        )
        return tf.squeeze(images, axis=0)

    def augment_images(self, images, transformations, **kwargs):
        original_shape = images.shape
        images = image_utils.transform_value_range(
            images, self.value_range, (0, 255), dtype=self.compute_dtype
        )
        inputs_for_aug_mix_single_image = {
            IMAGES: images,
            "transformations": transformations,
        }
        images = tf.map_fn(
            self.aug_mix_single_image,
            inputs_for_aug_mix_single_image,
            fn_output_signature=self.compute_dtype,
        )
        images = image_utils.transform_value_range(
            images, (0, 255), self.value_range, self.compute_dtype
        )
        images = tf.ensure_shape(images, shape=original_shape)
        return images

    def augment_labels(self, labels, transformations, **kwargs):
        return labels

    def aug_mix_single_image(self, inputs):
        image = inputs.get(IMAGES, None)
        transformation = inputs.get("transformations", None)

        chain_mixing_weights = tf.cast(
            transformation["chain_mixing_weights"], dtype=self.compute_dtype
        )
        weight_sample = tf.cast(
            transformation["weight_sample"], dtype=self.compute_dtype
        )

        result = tf.zeros_like(image, dtype=image.dtype)
        curr_chain = tf.constant([0], dtype=tf.int32)
        image, chain_mixing_weights, curr_chain, result = tf.while_loop(
            lambda image, chain_mixing_weights, curr_chain, result: tf.less(
                curr_chain, self.num_chains
            ),
            self.loop_on_width,
            [image, chain_mixing_weights, curr_chain, result],
        )
        result = weight_sample * image + (1 - weight_sample) * result
        return result

    def loop_on_width(self, image, chain_mixing_weights, curr_chain, result):
        image_aug = tf.identity(image)
        chain_depth = self._random_generator.random_uniform(
            shape=(),
            minval=self.chain_depth[0],
            maxval=self.chain_depth[1] + 1,
            dtype=tf.int32,
        )

        depth_level = tf.constant([0], dtype=tf.int32)
        depth_level, image_aug = tf.while_loop(
            lambda depth_level, image_aug: tf.less(depth_level, chain_depth),
            self.loop_on_depth,
            [depth_level, image_aug],
        )
        result += tf.gather(chain_mixing_weights, curr_chain) * image_aug
        curr_chain += 1
        return image, chain_mixing_weights, curr_chain, result

    def loop_on_depth(self, depth_level, image_aug):
        op_idx = self._random_generator.random_uniform(
            shape=(), minval=0, maxval=9, dtype=tf.int32
        )
        image_aug = self.apply_op(image_aug, op_idx)
        depth_level += 1
        return depth_level, image_aug

    def apply_op(self, image_aug, op_idx):
        augmented = image_aug
        augmented = tf.cond(
            op_idx == tf.constant([0], dtype=tf.int32),
            lambda: self.auto_contrast(augmented),
            lambda: augmented,
        )
        augmented = tf.cond(
            op_idx == tf.constant([1], dtype=tf.int32),
            lambda: self.equalize(augmented),
            lambda: augmented,
        )
        augmented = tf.cond(
            op_idx == tf.constant([2], dtype=tf.int32),
            lambda: self.posterize(augmented),
            lambda: augmented,
        )
        augmented = tf.cond(
            op_idx == tf.constant([3], dtype=tf.int32),
            lambda: self.rotate(augmented),
            lambda: augmented,
        )
        augmented = tf.cond(
            op_idx == tf.constant([4], dtype=tf.int32),
            lambda: self.solarize(augmented),
            lambda: augmented,
        )
        augmented = tf.cond(
            op_idx == tf.constant([5], dtype=tf.int32),
            lambda: self.shear(augmented, along_x=True),
            lambda: augmented,
        )
        augmented = tf.cond(
            op_idx == tf.constant([6], dtype=tf.int32),
            lambda: self.shear(augmented, along_x=False),
            lambda: augmented,
        )
        augmented = tf.cond(
            op_idx == tf.constant([7], dtype=tf.int32),
            lambda: self.translate(augmented, along_x=True),
            lambda: augmented,
        )
        augmented = tf.cond(
            op_idx == tf.constant([8], dtype=tf.int32),
            lambda: self.translate(augmented, along_x=False),
            lambda: augmented,
        )
        return augmented

    def posterize(self, image):
        ori_dtype = image.dtype
        bits = tf.cast(self.severity() * 3, tf.int32)
        shift = tf.cast(4 - bits + 1, tf.uint8)
        image = tf.cast(image, tf.uint8)
        image = tf.bitwise.left_shift(
            tf.bitwise.right_shift(image, shift), shift
        )
        image = tf.cast(image, dtype=ori_dtype)
        return image

    def rotate(self, image):
        angle = tf.expand_dims(
            self.severity(shape=(1,), dtype=tf.float32) * 30, axis=0
        )
        height = tf.expand_dims(tf.shape(image)[H_AXIS : H_AXIS + 1], axis=0)
        width = tf.expand_dims(tf.shape(image)[W_AXIS : W_AXIS + 1], axis=0)
        height = tf.cast(height, dtype=tf.float32)
        width = tf.cast(width, dtype=tf.float32)
        # tf.raw_ops.ImageProjectiveTransformV3 not support bfloat16
        if image.dtype == tf.bfloat16:
            image = tf.cast(image, dtype=tf.float32)
        image = image_utils.transform(
            tf.expand_dims(image, axis=0),
            image_utils.get_rotation_matrix(angle, height, width),
        )
        image = tf.squeeze(image, axis=0)
        return tf.cast(image, dtype=self.compute_dtype)

    def solarize(self, image):
        threshold = tf.cast(
            tf.cast(self.severity() * 255, tf.int32), image.dtype
        )
        image = tf.where(image < threshold, image, 255 - image)
        return image

    def shear(self, image, along_x=True):
        factor = tf.cast(self.severity() * 0.3, tf.float32)
        factor *= augmentation_utils.random_inversion(self._random_generator)
        if along_x:
            transform = tf.convert_to_tensor(
                [1.0, factor, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0]
            )
        else:
            transform = tf.convert_to_tensor(
                [1.0, 0.0, 0.0, factor, 1.0, 0.0, 0.0, 0.0]
            )
        # tf.raw_ops.ImageProjectiveTransformV3 not support bfloat16
        if image.dtype == tf.bfloat16:
            image = tf.cast(image, dtype=tf.float32)
        image = image_utils.transform(
            tf.expand_dims(image, axis=0),
            tf.expand_dims(transform, axis=0),
        )
        image = tf.squeeze(image, axis=0)
        return tf.cast(image, dtype=self.compute_dtype)

    def translate(self, image, along_x=True):
        shape = tf.cast(tf.shape(image), tf.float32)
        if along_x:
            size = shape[1]
        else:
            size = shape[0]
        factor = tf.cast(self.severity() * size / 3, tf.float32)
        factor *= augmentation_utils.random_inversion(self._random_generator)
        if along_x:
            transform = tf.convert_to_tensor(
                [1.0, 0.0, factor, 0.0, 1.0, 0.0, 0.0, 0.0]
            )
        else:
            transform = tf.convert_to_tensor(
                [1.0, 0.0, 0.0, 0.0, 1.0, factor, 0.0, 0.0]
            )
        # tf.raw_ops.ImageProjectiveTransformV3 not support bfloat16
        if image.dtype == tf.bfloat16:
            image = tf.cast(image, dtype=tf.float32)
        image = image_utils.transform(
            tf.expand_dims(image, axis=0),
            tf.expand_dims(transform, axis=0),
        )
        image = tf.squeeze(image, axis=0)
        return tf.cast(image, dtype=self.compute_dtype)

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "value_range": self.value_range,
                "severity": self.severity,
                "num_chains": self.num_chains,
                "chain_depth": self.chain_depth,
                "alpha": self.alpha,
                "seed": self.seed,
            }
        )
        return config



"""
Most of these codes come from KerasCV.
"""
import tensorflow as tf

from keras_aug.datapoints import bounding_box


class ToRaggedTest(tf.test.TestCase):
    def test_converts_to_ragged(self):
        bounding_boxes = {
            "boxes": tf.constant(
                [[[0, 0, 0, 0], [0, 0, 0, 0]], [[2, 3, 4, 5], [0, 1, 2, 3]]]
            ),
            "classes": tf.constant([[-1, -1], [-1, 1]]),
            "confidence": tf.constant([[0.5, 0.7], [0.23, 0.12]]),
        }

        bounding_boxes = bounding_box.to_ragged(bounding_boxes)

        self.assertEqual(bounding_boxes["boxes"][1].shape, [1, 4])
        self.assertEqual(bounding_boxes["classes"][1].shape, [1])
        self.assertEqual(bounding_boxes["confidence"][1].shape, [1])
        self.assertEqual(bounding_boxes["classes"][0].shape, [0])
        self.assertEqual(bounding_boxes["boxes"][0].shape, [0, 4])
        self.assertEqual(bounding_boxes["confidence"][0].shape, [0])

    def test_round_trip(self):
        original = {
            "boxes": tf.constant(
                [
                    [[0, 0, 0, 0], [-1, -1, -1, -1]],
                    [[-1, -1, -1, -1], [-1, -1, -1, -1]],
                ]
            ),
            "classes": tf.constant([[1, -1], [-1, -1]]),
            "confidence": tf.constant([[0.5, -1], [-1, -1]]),
        }

        bounding_boxes = bounding_box.to_ragged(original)
        bounding_boxes = bounding_box.to_dense(bounding_boxes, max_boxes=2)

        self.assertEqual(bounding_boxes["boxes"][1].shape, [2, 4])
        self.assertEqual(bounding_boxes["classes"][1].shape, [2])
        self.assertEqual(bounding_boxes["classes"][0].shape, [2])
        self.assertEqual(bounding_boxes["boxes"][0].shape, [2, 4])
        self.assertEqual(bounding_boxes["confidence"][0].shape, [2])

        self.assertAllEqual(bounding_boxes["boxes"], original["boxes"])
        self.assertAllEqual(bounding_boxes["classes"], original["classes"])
        self.assertAllEqual(
            bounding_boxes["confidence"], original["confidence"]
        )



import tensorflow as tf
from tensorflow import keras

from keras_aug.datapoints import image as image_utils
from keras_aug.layers.base.vectorized_base_random_layer import (
    VectorizedBaseRandomLayer,
)
from keras_aug.utils.augmentation import BATCHED
from keras_aug.utils.augmentation import H_AXIS
from keras_aug.utils.augmentation import LABELS
from keras_aug.utils.augmentation import SEGMENTATION_MASKS
from keras_aug.utils.augmentation import W_AXIS
from keras_aug.utils.distribution import stateless_random_beta


@keras.utils.register_keras_serializable(package="keras_aug")
class CutMix(VectorizedBaseRandomLayer):
    """CutMix implements the CutMix data augmentation technique.

    CutMix only supports dense images as inputs.

    Args:
        alpha (float, optional): The inverse scale parameter between 0 to +inf
            for the gamma distribution. This controls the shape of the
            distribution from which the smoothing values are sampled.
            Defaults to ``1.0``, which is a recommended value when training an
            ImageNet classification model.
        seed (int|float, optional): The random seed. Defaults to ``None``.

    References:
        - `CutMix <https://arxiv.org/abs/1905.04899>`_
        - `KerasCV <https://github.com/keras-team/keras-cv>`_
    """

    def __init__(
        self,
        alpha=1.0,
        seed=None,
        **kwargs,
    ):
        super().__init__(seed=seed, **kwargs)
        self.alpha = alpha
        self.seed = seed

        # set force_no_unwrap_ragged_image_call=True because CutMix needs
        # to process images in batch.
        self.force_no_unwrap_ragged_image_call = True

    def get_random_transformation_batch(
        self, batch_size, images=None, **kwargs
    ):
        height = tf.cast(tf.shape(images)[H_AXIS], dtype=tf.float32)
        width = tf.cast(tf.shape(images)[W_AXIS], dtype=tf.float32)
        permutation_order = self._random_generator.random_uniform(
            (batch_size,),
            minval=0,
            maxval=batch_size,
            dtype=tf.int32,
        )
        lambda_samples = stateless_random_beta(
            (batch_size,),
            seed_alpha=self._random_generator.make_seed_for_stateless_op(),
            seed_beta=self._random_generator.make_seed_for_stateless_op(),
            alpha=self.alpha,
            beta=self.alpha,
            dtype=tf.float32,
        )
        ratios = tf.math.sqrt(1.0 - lambda_samples)
        cut_heights = tf.cast(ratios * height, dtype=tf.int32)
        cut_widths = tf.cast(ratios * width, dtype=tf.int32)
        center_xs = self._random_generator.random_uniform(shape=(batch_size,))
        center_xs = tf.cast(center_xs * width, dtype=tf.int32)
        center_ys = self._random_generator.random_uniform(shape=(batch_size,))
        center_ys = tf.cast(center_ys * height, dtype=tf.int32)
        return {
            "permutation_order": permutation_order,
            "center_xs": center_xs,
            "center_ys": center_ys,
            "cut_heights": cut_heights,
            "cut_widths": cut_widths,
        }

    def augment_images(self, images, transformations, **kwargs):
        if isinstance(images, tf.RaggedTensor):
            raise ValueError(
                "CutMix expects dense images. Received: images type: "
                f"{type(images)}"
            )
        permutation_order = transformations["permutation_order"]
        center_xs = transformations["center_xs"]
        center_ys = transformations["center_ys"]
        cut_heights = transformations["cut_heights"]
        cut_widths = transformations["cut_widths"]
        images = image_utils.fill_rectangle(
            images,
            center_xs,
            center_ys,
            cut_widths,
            cut_heights,
            tf.gather(images, permutation_order),
        )
        return images

    def augment_labels(self, labels, transformations, images=None, **kwargs):
        labels = tf.cast(labels, dtype=self.compute_dtype)
        height = tf.cast(tf.shape(images)[H_AXIS], dtype=self.compute_dtype)
        width = tf.cast(tf.shape(images)[W_AXIS], dtype=self.compute_dtype)
        permutation_order = transformations["permutation_order"]
        cutmix_labels = tf.gather(labels, permutation_order)

        cut_heights = tf.cast(
            transformations["cut_heights"], dtype=self.compute_dtype
        )
        cut_widths = tf.cast(
            transformations["cut_widths"], dtype=self.compute_dtype
        )
        bounding_box_area = cut_heights * cut_widths
        lambda_sample = 1.0 - bounding_box_area / (height * width)
        lambda_sample = tf.reshape(lambda_sample, [-1, 1])

        labels = lambda_sample * labels + (1.0 - lambda_sample) * cutmix_labels
        return labels

    def augment_segmentation_masks(
        self, segmentation_masks, transformations, **kwargs
    ):
        if isinstance(segmentation_masks, tf.RaggedTensor):
            raise ValueError(
                "CutMix expects dense segmentation_masks. Received: "
                f"segmentation_masks type: {type(segmentation_masks)}"
            )
        permutation_order = transformations["permutation_order"]
        center_xs = transformations["center_xs"]
        center_ys = transformations["center_ys"]
        cut_heights = transformations["cut_heights"]
        cut_widths = transformations["cut_widths"]
        segmentation_masks = image_utils.fill_rectangle(
            segmentation_masks,
            center_xs,
            center_ys,
            cut_widths,
            cut_heights,
            tf.gather(segmentation_masks, permutation_order),
        )
        return segmentation_masks

    def _batch_augment(self, inputs):
        self._validate_inputs(inputs)
        return super()._batch_augment(inputs)

    def call(self, inputs):
        _, metadata = self._format_inputs(inputs)
        if metadata[BATCHED] is not True:
            raise ValueError(
                "CutMix received a single image to `call`. The "
                "layer relies on combining multiple examples, and as such "
                "will not behave as expected. Please call the layer with 2 "
                "or more samples."
            )
        return super().call(inputs=inputs)

    def _validate_inputs(self, inputs):
        labels = inputs.get(LABELS, None)
        segmentation_masks = inputs.get(SEGMENTATION_MASKS, None)
        if labels is None and segmentation_masks is None:
            raise ValueError(
                "CutMix expects `labels` or `segmentation_masks` to be present "
                "in its inputs. "
                "For example, `cut_mix({'images': images, 'labels': labels})`."
            )

    def get_config(self):
        config = super().get_config()
        config.update({"alpha": self.alpha, "seed": self.seed})
        return config



import math

import tensorflow as tf
from tensorflow import keras

from keras_aug.datapoints import image as image_utils
from keras_aug.layers.base.vectorized_base_random_layer import (
    VectorizedBaseRandomLayer,
)
from keras_aug.utils import augmentation as augmentation_utils
from keras_aug.utils.augmentation import H_AXIS
from keras_aug.utils.augmentation import IMAGES
from keras_aug.utils.augmentation import W_AXIS


@keras.utils.register_keras_serializable(package="keras_aug")
class RandomGridMask(VectorizedBaseRandomLayer):
    """RandomGridMask performs the Grid Mask operation on input images.

    Args:
        size_factor (float|Sequence[float]|keras_aug.FactorSampler, optional):
            The relative size for grid masks. When represented as a single
            float, the factor will be picked between ``[0.0, 0.0 + upper]``.
            Represented as d1, d2 in the paper. Defaults to
            ``(96/224, 224/224)`` which is for ImageNet classification model.
            For COCO object detection, it is set to ``(0.01, 1.0)``
        ratio_factor (float|Sequence[float]|keras_aug.FactorSampler, optional):
            The ratio from spacings to grid masks. When represented as a single
            float, the factor will be picked between ``[0.0, 0.0 + upper]``.
            Represented as ratio in the paper. Lower values make the grid size
            smaller, and higher values make the grid mask large. ``0.5``
            indicates that grid and spacing will be of equal size. Defaults to
            ``(0.6, 0.6)`` which is for ImageNet classification model. For COCO
            object detection, it is set to ``(0.5, 0.5)``
        rotation_factor (float|Sequence[float]|keras_aug.FactorSampler, optional):
            The range of the degree that will be used to rotate the grid_mask.
            When represented as a single float, the factor will be picked
            between ``[0.0 - lower, 0.0 + upper]``. A positive value means
            rotating counter clock-wise, while a negative value means
            clock-wise. Defaults to ``(-180, 180)`` which is for ImageNet
            classification model. For COCO object detection, it is set to
            ``(0, 0)``.
        fill_mode (str, optional): The fill mode. Supported values:
            ``"constant", "gaussian_noise", "random"``. Defaults to
            ``"constant"``.
        fill_value (int|float, optional): The value to be filled inside the
            gridblock when ``fill_mode="constant"``. Defaults to ``0``.
        seed (int|float, optional): The random seed. Defaults to ``None``.

    References:
        - `GridMask <https://arxiv.org/abs/2001.04086>`_
        - `GridMask Official Repo <https://github.com/dvlab-research/GridMask>`_
        - `KerasCV <https://github.com/keras-team/keras-cv>`_
    """  # noqa: E501

    def __init__(
        self,
        size_factor=(96.0 / 224.0, 224.0 / 224.0),
        ratio_factor=(0.6, 0.6),
        rotation_factor=(-180, 180),
        fill_mode="constant",
        fill_value=0.0,
        seed=None,
        **kwargs,
    ):
        super().__init__(seed=seed, **kwargs)
        if isinstance(size_factor, (int, float)):
            size_factor = (0, size_factor)
        # d1 cannot be too small, the computation of grid masks might be
        # expensive
        self.size_factor = augmentation_utils.parse_factor(
            size_factor, min_value=0.01, seed=seed
        )
        if isinstance(ratio_factor, (int, float)):
            ratio_factor = (0, ratio_factor)
        self.ratio_factor = augmentation_utils.parse_factor(
            ratio_factor, seed=seed
        )
        if isinstance(rotation_factor, (int, float)):
            rotation_factor = (-rotation_factor, rotation_factor)
        self.rotation_factor = augmentation_utils.parse_factor(
            rotation_factor, min_value=-180, max_value=180, seed=seed
        )
        self._check_parameter_values(fill_mode)
        self.fill_mode = fill_mode
        self.fill_value = fill_value
        self.seed = seed

        # decide whether to enable the rotation
        self._enable_rotation = augmentation_utils.is_factor_working(
            self.rotation_factor, not_working_value=0.0
        )

    def _check_parameter_values(self, fill_mode):
        if fill_mode not in ["constant", "gaussian_noise", "random"]:
            raise ValueError(
                '`fill_mode` should be "constant", '
                '"gaussian_noise", or "random". Got '
                f"`fill_mode`={fill_mode}"
            )

    def get_random_transformation_batch(
        self, batch_size, images=None, **kwargs
    ):
        # all operations run in tf.float32 to avoid numerical issue
        heights, widths = augmentation_utils.get_images_shape(
            images, dtype=tf.float32
        )

        # mask side length
        input_diagonal_lens = tf.sqrt(tf.square(widths) + tf.square(heights))
        mask_side_lens = tf.math.ceil(input_diagonal_lens)

        # grid unit size
        smaller_side_lens = tf.where(heights < widths, heights, widths)
        unit_ratios = self.size_factor(shape=(batch_size, 1))
        unit_sizes = unit_ratios * smaller_side_lens
        unit_sizes = tf.maximum(unit_sizes, 2)  # prevent too small units
        ratios = self.ratio_factor(shape=(batch_size, 1))
        rectangle_side_lens = ratios * unit_sizes

        # sample x and y offset for grid units randomly between 0 and unit_size
        delta_xs = self._random_generator.random_uniform(shape=(batch_size, 1))
        delta_ys = self._random_generator.random_uniform(shape=(batch_size, 1))
        delta_xs = delta_xs * unit_sizes
        delta_ys = delta_ys * unit_sizes

        # randomly rotate mask
        angles = self.rotation_factor(shape=(batch_size, 1))
        angles = angles / 360.0 * 2.0 * math.pi

        return {
            "mask_side_lens": mask_side_lens,
            "rectangle_side_lens": rectangle_side_lens,
            "unit_sizes": unit_sizes,
            "delta_xs": delta_xs,
            "delta_ys": delta_ys,
            "angles": angles,
        }

    def augment_ragged_image(self, image, transformation, **kwargs):
        image = tf.expand_dims(image, axis=0)
        transformation = augmentation_utils.expand_dict_dims(
            transformation, axis=0
        )
        image = self.augment_images(
            images=image, transformations=transformation, **kwargs
        )
        return tf.squeeze(image, axis=0)

    def augment_images(self, images, transformations, **kwargs):
        masks = tf.map_fn(
            self.compute_grid_mask_single_image,
            {IMAGES: images, **transformations},
            fn_output_signature=tf.bool,
        )
        masks = tf.expand_dims(masks, axis=-1)
        # masks (batch_size, height, width, 1)

        if self._enable_rotation:
            angles = transformations.get("angles", None)
            heights, widths = augmentation_utils.get_images_shape(
                images, dtype=tf.float32
            )
            rotation_matrixes = image_utils.get_rotation_matrix(
                angles, heights, widths
            )
            masks = image_utils.transform(
                tf.cast(masks, dtype=tf.float32),
                rotation_matrixes,
                fill_mode="constant",
                fill_value=0,
                interpolation="nearest",
            )
        # center crop mask
        height = tf.shape(images)[H_AXIS]
        width = tf.shape(images)[W_AXIS]
        h_diff = tf.shape(masks)[H_AXIS] - height
        w_diff = tf.shape(masks)[W_AXIS] - width
        h_start = tf.cast(h_diff / 2, tf.int32)
        w_start = tf.cast(w_diff / 2, tf.int32)
        masks = tf.image.crop_to_bounding_box(
            masks, h_start, w_start, height, width
        )
        # convert back to boolean mask
        masks = tf.cast(masks, tf.bool)
        # compute fill
        if self.fill_mode == "constant":
            fill_value = tf.fill(tf.shape(images), self.fill_value)
            fill_value = tf.cast(fill_value, dtype=self.compute_dtype)
        elif self.fill_mode == "gaussian_noise":
            fill_value = self._random_generator.random_normal(
                shape=tf.shape(images), dtype=self.compute_dtype
            )
        else:
            fill_value = self._random_generator.random_uniform(
                shape=tf.shape(images), dtype=self.compute_dtype
            )
        return tf.where(masks, fill_value, images)

    def augment_labels(self, labels, transformations, **kwargs):
        return labels

    def augment_bounding_boxes(self, bounding_boxes, transformations, **kwargs):
        return bounding_boxes

    def compute_grid_mask_single_image(self, inputs):
        mask_side_len = inputs.get("mask_side_lens", None)[0]
        rectangle_side_len = inputs.get("rectangle_side_lens", None)[0]
        unit_size = inputs.get("unit_sizes", None)[0]
        delta_x = inputs.get("delta_xs", None)[0]
        delta_y = inputs.get("delta_ys", None)[0]

        # grid size (number of diagonal units in grid)
        grid_size = tf.cast(mask_side_len // unit_size + 1, dtype=tf.int32)
        grid_size_range = tf.range(1, grid_size + 1)
        # diagonal corner coordinates
        unit_size_range = (
            tf.cast(grid_size_range, dtype=unit_size.dtype) * unit_size
        )
        x1 = unit_size_range - delta_x
        x0 = x1 - rectangle_side_len
        y1 = unit_size_range - delta_y
        y0 = y1 - rectangle_side_len
        # compute grid coordinates
        x0, y0 = tf.meshgrid(x0, y0)
        x1, y1 = tf.meshgrid(x1, y1)
        # flatten mesh grid
        x0 = tf.reshape(x0, [-1])
        y0 = tf.reshape(y0, [-1])
        x1 = tf.reshape(x1, [-1])
        y1 = tf.reshape(y1, [-1])
        # convert coordinates to mask
        # corners must be tf.float32 for fill_utils.corners_to_mask
        corners = tf.cast(tf.stack([x0, y0, x1, y1], axis=-1), dtype=tf.float32)
        mask_side_len = tf.cast(mask_side_len, dtype=tf.int32)
        rectangle_masks = image_utils.corners_to_mask(
            corners, mask_shape=(mask_side_len, mask_side_len)
        )
        grid_mask = tf.reduce_any(rectangle_masks, axis=0)
        return grid_mask

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "size_factor": self.size_factor,
                "ratio_factor": self.ratio_factor,
                "rotation_factor": self.rotation_factor,
                "fill_mode": self.fill_mode,
                "fill_value": self.fill_value,
                "seed": self.seed,
            }
        )
        return config



try:
    import keras.src.backend

    KerasRandomGenerator = keras.src.backend.RandomGenerator

except ModuleNotFoundError:
    """Backwards compatibility for keras<=2.12"""
    import keras.backend

    keras.backend.RandomGenerator
    KerasRandomGenerator = keras.backend.RandomGenerator



import tensorflow as tf
from tensorflow import keras

from keras_aug.datapoints import image as image_utils
from keras_aug.layers.base.vectorized_base_random_layer import (
    VectorizedBaseRandomLayer,
)
from keras_aug.utils import augmentation as augmentation_utils


@keras.utils.register_keras_serializable(package="keras_aug")
class RandomErase(VectorizedBaseRandomLayer):
    """Randomly erase rectangles from images and fill them.

    Args:
        area_factor (float|Sequence[float]|keras_aug.FactorSampler, optional): The range
            of the area factor that controls the area of the erasing. When
            represented as a single float, the factor will be picked between
            ``[0.0, 0.0 + upper]``. ``0.0`` means the rectangle will be of size
            0% of the image area. ``0.1`` means the rectangle will have a size
            of 10% of the image area. Defaults to ``(0.02, 0.4)``
        aspect_ratio_factor (float|Sequence[float]|keras_aug.FactorSampler, optional): The
            range of the aspect ratio factor that controls the aspect ratio of
            the erasing. When represented as a single float, the factor will be
            picked between ``[1.0 - lower, 1.0 + upper]``. ``1.0`` means the
            erasing will be square. Defaults to ``(0.3, 1.0 / 0.3)``.
        fill_mode (str, optional): Pixels inside the erasing are filled
            according to the given mode. Supported values:
            ``"constant", "gaussian_noise"``. Defaults to ``"constant"``.
        fill_value (tuple(float), optional): The values to be filled in the
            erasing when ``fill_mode="constant"``. Defaults to
            ``(125, 123, 114)`` which is the means of ImageNet.
        seed (int|float, optional): The random seed. Defaults to ``None``.

    References:
        - `Random Erasing <https://arxiv.org/abs/1708.04896v2>`_
        - `Random Erasing Official Repo <https://github.com/zhunzhong07/Random-Erasing>`_
        - `Tensorflow Model augment <https://github.com/tensorflow/models/blob/v2.12.0/official/vision/ops/augment.py>`_
        - `KerasCV <https://github.com/keras-team/keras-cv>`_
    """  # noqa: E501

    def __init__(
        self,
        area_factor=(0.02, 0.4),
        aspect_ratio_factor=(0.3, 1.0 / 0.3),
        fill_mode="constant",
        fill_value=(125, 123, 114),
        seed=None,
        **kwargs,
    ):
        super().__init__(seed=seed, **kwargs)
        if isinstance(area_factor, (int, float)):
            area_factor = (0, area_factor)
        self.area_factor = augmentation_utils.parse_factor(
            area_factor, seed=seed
        )
        if isinstance(aspect_ratio_factor, (int, float)):
            aspect_ratio_factor = (
                1.0 - aspect_ratio_factor,
                1.0 + aspect_ratio_factor,
            )
        self.aspect_ratio_factor = augmentation_utils.parse_factor(
            aspect_ratio_factor, max_value=None, seed=seed
        )
        if fill_mode not in ["gaussian_noise", "constant"]:
            raise ValueError(
                '`fill_mode` should be "gaussian_noise" '
                f'or "constant". Got `fill_mode`={fill_mode}'
            )
        self.fill_mode = fill_mode
        self.fill_value = tf.convert_to_tensor(
            fill_value, dtype=self.compute_dtype
        )
        self.seed = seed

        self.max_attemp = 10

    def get_random_transformation_batch(
        self, batch_size, images=None, **kwargs
    ):
        heights, widths = augmentation_utils.get_images_shape(
            images, dtype=tf.float32
        )
        areas = heights * widths

        is_success = False
        for _ in range(self.max_attemp):
            if not is_success:
                erasing_areas = self.area_factor(shape=(batch_size, 1))
                erasing_areas = erasing_areas * areas
                erasing_aspect_ratios = self.aspect_ratio_factor(
                    shape=(batch_size, 1)
                )
                erasing_heights = tf.round(
                    tf.sqrt(erasing_areas * erasing_aspect_ratios)
                )
                erasing_widths = tf.round(
                    tf.sqrt(erasing_areas / erasing_aspect_ratios)
                )
                if tf.reduce_all(
                    tf.less(erasing_widths, widths)
                ) and tf.reduce_all(tf.less(erasing_heights, heights)):
                    is_success = True

        center_xs = self._random_generator.random_uniform(
            shape=(batch_size, 1), minval=0, maxval=1, dtype=tf.float32
        )
        center_ys = self._random_generator.random_uniform(
            shape=(batch_size, 1), minval=0, maxval=1, dtype=tf.float32
        )
        center_xs = tf.round(center_xs * (widths - erasing_widths))
        center_xs = tf.cast(center_xs + erasing_widths / 2, dtype=tf.int32)
        center_ys = tf.round(center_ys * (heights - erasing_heights))
        center_ys = tf.cast(center_ys + erasing_heights / 2, dtype=tf.int32)

        return {
            "center_xs": center_xs,
            "center_ys": center_ys,
            "erasing_heights": tf.cast(erasing_heights, dtype=tf.int32),
            "erasing_widths": tf.cast(erasing_widths, dtype=tf.int32),
        }

    def augment_ragged_image(self, image, transformation, **kwargs):
        image = tf.expand_dims(image, axis=0)
        transformation = augmentation_utils.expand_dict_dims(
            transformation, axis=0
        )
        image = self.augment_images(
            images=image, transformations=transformation, **kwargs
        )
        return tf.squeeze(image, axis=0)

    def augment_images(self, images, transformations, **kwargs):
        images = tf.cast(images, dtype=self.compute_dtype)
        center_xs = transformations["center_xs"]
        center_ys = transformations["center_ys"]
        erasing_heights = transformations["erasing_heights"]
        erasing_widths = transformations["erasing_widths"]
        rectangle_fills = self.compute_rectangle_fills(images)
        images = image_utils.fill_rectangle(
            images,
            center_xs[..., 0],
            center_ys[..., 0],
            erasing_widths[..., 0],
            erasing_heights[..., 0],
            rectangle_fills,
        )
        return images

    def augment_labels(self, labels, transformations, **kwargs):
        return labels

    def compute_rectangle_fills(self, inputs):
        if self.fill_mode == "constant":
            fills = tf.ones(tf.shape(inputs), dtype=self.compute_dtype)
            fills = (
                fills * self.fill_value[tf.newaxis, tf.newaxis, tf.newaxis, ...]
            )
        else:
            # gaussian noise
            fills = tf.random.normal(tf.shape(inputs), dtype=self.compute_dtype)
        return fills

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "area_factor": self.area_factor,
                "aspect_ratio_factor": self.aspect_ratio_factor,
                "fill_mode": self.fill_mode,
                "fill_value": self.fill_value,
                "seed": self.seed,
            }
        )
        return config



from functools import partial

import tensorflow as tf
from tensorflow import keras

from keras_aug.datapoints import bounding_box
from keras_aug.layers.base.vectorized_base_random_layer import (
    VectorizedBaseRandomLayer,
)
from keras_aug.utils import augmentation as augmentation_utils
from keras_aug.utils.augmentation import BOUNDING_BOXES
from keras_aug.utils.augmentation import CUSTOM_ANNOTATIONS
from keras_aug.utils.augmentation import IMAGES
from keras_aug.utils.augmentation import KEYPOINTS
from keras_aug.utils.augmentation import LABELS
from keras_aug.utils.augmentation import SEGMENTATION_MASKS


@keras.utils.register_keras_serializable(package="keras_aug")
class RandomChoice(VectorizedBaseRandomLayer):
    """RandomChoice constructs a pipeline based on provided arguments.

    The implemented policy does the following: for each input provided in
    `call`(), the policy selects a random layer from the provided list of
    `layers`. It then calls the `layer()` on the inputs.

    Notes:
        The shape and type of the outputs must be the same of the inputs.

    Args:
        layers (list(VectorizedBaseRandomLayer|keras.Layer|keras.Sequential)): The list
            of the layers that will be picked randomly for the pipeline.
        batchwise (bool, optional): Whether to pass entire batches to the
            underlying layer. When set to ``True``, each batch is passed to a
            single layer, instead of each sample to an independent layer. This
            is useful when using ``MixUp()``, ``CutMix()``, ``Mosaic()``, etc.
            Defaults to ``False``.
        seed (int|float, optional): The random seed. Defaults to ``None``.

    References:
        - `KerasCV <https://github.com/keras-team/keras-cv>`_
    """  # noqa: E501

    def __init__(self, layers, batchwise=False, seed=None, **kwargs):
        super().__init__(seed=seed, **kwargs)
        self.layers = layers
        self.batchwise = batchwise
        self.seed = seed

    def get_random_transformation_batch(self, batch_size, **kwargs):
        selected_op_idx = self._random_generator.random_uniform(
            shape=(batch_size,),
            minval=0,
            maxval=len(self.layers),
            dtype=tf.int32,
        )
        return selected_op_idx

    def compute_inputs_signature(self, inputs):
        fn_output_signature = {}
        if IMAGES in inputs:
            if isinstance(inputs[IMAGES], tf.Tensor):
                fn_output_signature[IMAGES] = tf.TensorSpec(
                    inputs[IMAGES].shape[1:], self.compute_dtype
                )
            else:
                fn_output_signature[IMAGES] = tf.RaggedTensorSpec(
                    shape=inputs[IMAGES].shape[1:],
                    ragged_rank=1,
                    dtype=self.compute_dtype,
                )
        if LABELS in inputs:
            fn_output_signature[LABELS] = tf.TensorSpec(
                inputs[LABELS].shape[1:], self.compute_dtype
            )
        if BOUNDING_BOXES in inputs:
            fn_output_signature[BOUNDING_BOXES] = {
                "boxes": tf.RaggedTensorSpec(
                    shape=[None, 4],
                    ragged_rank=1,
                    dtype=self.compute_dtype,
                ),
                "classes": tf.RaggedTensorSpec(
                    shape=[None], ragged_rank=0, dtype=self.compute_dtype
                ),
            }
        if SEGMENTATION_MASKS in inputs:
            if isinstance(inputs[SEGMENTATION_MASKS], tf.Tensor):
                fn_output_signature[SEGMENTATION_MASKS] = tf.TensorSpec(
                    inputs[SEGMENTATION_MASKS].shape[1:], self.compute_dtype
                )
            else:
                fn_output_signature[SEGMENTATION_MASKS] = tf.RaggedTensorSpec(
                    shape=inputs[SEGMENTATION_MASKS].shape[1:],
                    ragged_rank=1,
                    dtype=self.compute_dtype,
                )
        if KEYPOINTS in inputs:
            if isinstance(inputs[KEYPOINTS], tf.Tensor):
                fn_output_signature[KEYPOINTS] = tf.TensorSpec(
                    inputs[KEYPOINTS].shape[1:], self.compute_dtype
                )
            else:
                fn_output_signature[KEYPOINTS] = tf.RaggedTensorSpec(
                    shape=inputs[KEYPOINTS].shape[1:],
                    ragged_rank=1,
                    dtype=self.compute_dtype,
                )
        if CUSTOM_ANNOTATIONS in inputs:
            raise NotImplementedError()
        return fn_output_signature

    def _batch_augment(self, inputs):
        images = inputs.get(IMAGES, None)
        batch_size = tf.shape(images)[0]
        transformations = self.get_random_transformation_batch(batch_size)
        if self.batchwise:
            selected_op_idx = transformations[0]
            result = self.augment(
                {"inputs": inputs, "transformations": selected_op_idx}
            )
        else:
            bounding_boxes = inputs.get(BOUNDING_BOXES, None)
            # make bounding_boxes to dense first
            if bounding_boxes is not None:
                ori_bbox_info = bounding_box.validate_format(bounding_boxes)
                inputs[BOUNDING_BOXES] = bounding_box.to_dense(bounding_boxes)

            inputs_for_random_choice_single_input = {
                "inputs": inputs,
                "transformations": transformations,
            }
            result = tf.map_fn(
                self.augment,
                inputs_for_random_choice_single_input,
                fn_output_signature=self.compute_inputs_signature(inputs),
            )

            bounding_boxes = result.get(BOUNDING_BOXES, None)
            if bounding_boxes is not None:
                if ori_bbox_info["ragged"]:
                    bounding_boxes = bounding_box.to_ragged(bounding_boxes)
                else:
                    bounding_boxes = bounding_box.to_dense(bounding_boxes)
                result[BOUNDING_BOXES] = bounding_boxes
        return result

    def augment(self, inputs):
        input = inputs.get("inputs")
        selected_op_idx = inputs.get("transformations")
        # construct branch_fns
        branch_fns = {}
        for idx, layer in enumerate(self.layers):
            branch_fns[idx] = partial(layer, input)
        # augment
        result = tf.switch_case(selected_op_idx, branch_fns=branch_fns)
        if BOUNDING_BOXES in result:
            result[BOUNDING_BOXES] = bounding_box.to_ragged(
                result[BOUNDING_BOXES], dtype=self.compute_dtype
            )
        result = augmentation_utils.cast_to(result, dtype=self.compute_dtype)
        return result

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "layers": self.layers,
                "batchwise": self.batchwise,
                "seed": self.seed,
            }
        )
        return config



import tensorflow as tf

from keras_aug import layers


class IdentityTest(tf.test.TestCase):
    def test_all_inputs_intact(self):
        images = tf.random.uniform(shape=(2, 4, 4, 3))
        labels = tf.random.uniform(
            shape=(2, 1), minval=0, maxval=10, dtype=tf.int32
        )
        bounding_boxes = {
            "boxes": tf.ragged.stack(
                [
                    tf.ones((2, 4)),
                    tf.ones((3, 4)),
                ]
            ),
            "classes": tf.ragged.stack(
                [
                    tf.ones((2,)),
                    tf.ones((3,)),
                ]
            ),
        }
        segmentation_masks = tf.random.uniform(
            shape=(2, 4, 4, 3), minval=0, maxval=10, dtype=tf.int32
        )
        inputs = {
            "images": images,
            "labels": labels,
            "bounding_boxes": bounding_boxes,
            "segmentation_masks": segmentation_masks,
        }
        layer = layers.Identity()

        outputs = layer(inputs)

        for key in inputs:
            if key == "bounding_boxes":
                self.assertAllEqual(outputs[key]["boxes"], inputs[key]["boxes"])
                self.assertAllEqual(
                    outputs[key]["classes"], inputs[key]["classes"]
                )
            else:
                self.assertAllEqual(outputs[key], inputs[key])



import tensorflow as tf
from absl.testing import parameterized
from tensorflow import keras

from keras_aug import layers


class AddOneToInputs(keras.layers.Layer):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.call_counter = tf.Variable(initial_value=0)

    def call(self, inputs):
        result = inputs.copy()
        result["images"] = inputs["images"] + 1
        self.call_counter.assign_add(1)
        return result


class RandomAugmentationPipelineTest(tf.test.TestCase, parameterized.TestCase):
    def test_calls_layer_augmentation_per_image(self):
        layer = AddOneToInputs()
        pipeline = layers.RandomChoice(layers=[layer])
        xs = tf.random.uniform((2, 5, 5, 3), 0, 100, dtype=tf.float32)
        os = pipeline(xs)

        self.assertAllClose(xs + 1, os)

    def test_calls_layer_augmentation_in_graph(self):
        layer = AddOneToInputs()
        pipeline = layers.RandomChoice(layers=[layer])

        @tf.function()
        def call_pipeline(xs):
            return pipeline(xs)

        xs = tf.random.uniform((2, 5, 5, 3), 0, 100, dtype=tf.float32)
        os = call_pipeline(xs)

        self.assertAllClose(xs + 1, os)

    def test_batchwise(self):
        layer = AddOneToInputs()
        pipeline = layers.RandomChoice(layers=[layer], batchwise=True)
        xs = tf.random.uniform((4, 5, 5, 3), 0, 100, dtype=tf.float32)
        os = pipeline(xs)

        self.assertAllClose(xs + 1, os)
        # Ensure the layer is only called once for the entire batch
        self.assertEqual(layer.call_counter, 1)

    def test_works_with_cutmix_mixup(self):
        pipeline = layers.RandomChoice(
            layers=[layers.CutMix(), layers.MixUp()], batchwise=True
        )
        xs = {
            "images": tf.random.uniform((4, 5, 5, 3), 0, 100, dtype=tf.float32),
            "labels": tf.random.uniform((4, 10), 0, 1, dtype=tf.float32),
        }
        pipeline(xs)

    def test_calls_layer_augmentation_single_image(self):
        layer = AddOneToInputs()
        pipeline = layers.RandomChoice(layers=[layer])
        xs = tf.random.uniform((5, 5, 3), 0, 100, dtype=tf.float32)
        os = pipeline(xs)

        self.assertAllClose(xs + 1, os)

    def test_calls_choose_one_layer_augmentation(self):
        batch_size = 10
        pipeline = layers.RandomChoice(
            layers=[AddOneToInputs(), AddOneToInputs()]
        )
        xs = tf.random.uniform((batch_size, 5, 5, 3), 0, 100, dtype=tf.float32)
        os = pipeline(xs)

        self.assertAllClose(xs + 1, os)

        total_calls = (
            pipeline.layers[0].call_counter + pipeline.layers[1].call_counter
        )
        self.assertEqual(total_calls, batch_size)



import tensorflow as tf
from tensorflow import keras

from keras_aug.datapoints import image as image_utils
from keras_aug.layers.base.vectorized_base_random_layer import (
    VectorizedBaseRandomLayer,
)
from keras_aug.utils import augmentation as augmentation_utils


@keras.utils.register_keras_serializable(package="keras_aug")
class RandomHSV(VectorizedBaseRandomLayer):
    """Randomly adjusts the hue, saturation and value on given images.

    This layer will randomly increase/reduce the hue, saturation and value for
    the input RGB images. The image hue, saturation and value is adjusted by
    converting the image(s) to HSV and rotating the hue channel (H) by hue
    factor, multiplying the saturation channel (S) by saturation factor and
    multiplying the value channel (V) by value factor. The image is then
    converted back to RGB.

    Args:
        value_range (Sequence[int|float]): The range of values the incoming
            images will have. This is typically either ``[0, 1]`` or
            ``[0, 255]`` depending on how your preprocessing pipeline is set up.
        hue_factor (float|Sequence[float]|keras_aug.FactorSampler, optional): The range of
            the hue factor. When represented as a single float, the factor will
            be picked between ``[0.5 - lower, 0.5 + upper]``. ``0.0`` means no
            shift. ``-0.5`` or ``0.5`` gives an image with complementary colors.
            Defaults to ``None``.
        saturation_factor (float|Sequence[float]|keras_aug.FactorSampler, optional): The
            range of the saturation factor. When represented as a single float,
            the factor will be picked between ``[1.0 - lower, 1.0 + upper]``.
            ``1.0`` will give the original image. ``0.0`` makes the image to be
            fully grayscale. ``2.0`` will enhance the saturation by a factor of
            2. Defaults to ``None``.
        value_factor (float|Sequence[float]|keras_aug.FactorSampler, optional): The range
            of the value factor. When represented as a single float,
            the factor will be picked between ``[1.0 - lower, 1.0 + upper]``.
            ``1.0`` will give the original image. ``0.0`` makes the image to be
            zero values. ``2.0`` will enhance the value by a factor of 2.
            Defaults to ``None``.
        seed (int|float, optional): The random seed. Defaults to
            ``None``.

    References:
        - `KerasCV <https://github.com/keras-team/keras-cv>`_
    """  # noqa: E501

    def __init__(
        self,
        value_range,
        hue_factor=None,
        saturation_factor=None,
        value_factor=None,
        seed=None,
        **kwargs,
    ):
        super().__init__(seed=seed, **kwargs)
        self.hue_factor = None
        self.saturation_factor = None
        self.value_factor = None
        if hue_factor is not None:
            self.hue_factor = augmentation_utils.parse_factor(
                hue_factor,
                min_value=-0.5,
                max_value=0.5,
                center_value=0,
                seed=seed,
            )
        if saturation_factor is not None:
            self.saturation_factor = augmentation_utils.parse_factor(
                saturation_factor, max_value=None, center_value=1, seed=seed
            )
        if value_factor is not None:
            self.value_factor = augmentation_utils.parse_factor(
                value_factor, max_value=None, center_value=1, seed=seed
            )
        self.value_range = value_range
        self.seed = seed

        # decide whether to enable the augmentation
        self._enable_hue = augmentation_utils.is_factor_working(
            self.hue_factor, not_working_value=0.0
        )
        self._enable_saturation = augmentation_utils.is_factor_working(
            self.saturation_factor, not_working_value=1.0
        )
        self._enable_value = augmentation_utils.is_factor_working(
            self.value_factor, not_working_value=1.0
        )

    def get_random_transformation_batch(self, batch_size, **kwargs):
        factor_shape = (batch_size, 1, 1, 1)
        # dummy
        hue_factors = tf.zeros(factor_shape)
        saturation_factors = tf.zeros(factor_shape)
        value_factors = tf.zeros(factor_shape)
        if self._enable_hue:
            hue_factors = self.hue_factor(
                shape=factor_shape, dtype=self.compute_dtype
            )
        if self._enable_saturation:
            saturation_factors = self.saturation_factor(
                shape=factor_shape, dtype=self.compute_dtype
            )
        if self._enable_value:
            value_factors = self.value_factor(
                shape=factor_shape, dtype=self.compute_dtype
            )
        return {
            "hue_factors": hue_factors,
            "saturation_factors": saturation_factors,
            "value_factors": value_factors,
        }

    def augment_ragged_image(self, image, transformation, **kwargs):
        images = tf.expand_dims(image, axis=0)
        transformations = augmentation_utils.expand_dict_dims(
            transformation, axis=0
        )
        images = self.augment_images(
            images=images, transformations=transformations, **kwargs
        )
        return tf.squeeze(images, axis=0)

    def augment_images(self, images, transformations, **kwargs):
        images = image_utils.transform_value_range(
            images, self.value_range, (0, 255), dtype=self.compute_dtype
        )
        if self._enable_hue:
            images = self.adjust_hue(images, transformations)
        if self._enable_saturation:
            images = self.adjust_saturation(images, transformations)
        if self._enable_value:
            images = self.adjust_value(images, transformations)
        images = image_utils.transform_value_range(
            images, (0, 255), self.value_range, dtype=self.compute_dtype
        )
        return images

    def augment_labels(self, labels, transformations, **kwargs):
        return labels

    def augment_bounding_boxes(self, bounding_boxes, transformations, **kwargs):
        return bounding_boxes

    def augment_segmentation_masks(
        self, segmentation_masks, transformations, **kwargs
    ):
        return segmentation_masks

    def augment_keypoints(self, keypoints, transformations, **kwargs):
        return keypoints

    def adjust_hue(self, images, transformations):
        # cast to float32 to avoid numerical issue
        images = tf.image.rgb_to_hsv(tf.cast(images, dtype=tf.float32))
        hue_factors = tf.cast(transformations["hue_factors"], dtype=tf.float32)
        h_channels = tf.math.floormod(images[..., 0:1] + hue_factors, 1.0)
        images = tf.concat(
            [h_channels, images[..., 1:2], images[..., 2:3]], axis=-1
        )
        images = tf.image.hsv_to_rgb(images)
        images = tf.clip_by_value(images, 0, 255)
        return tf.cast(images, dtype=self.compute_dtype)

    def adjust_saturation(self, images, transformations):
        saturation_factors = transformations["saturation_factors"]
        degenerates = image_utils.rgb_to_grayscale(images)
        images = image_utils.blend(
            degenerates, images, saturation_factors, (0, 255)
        )
        return images

    def adjust_value(self, images, transformations):
        value_factors = transformations["value_factors"]
        images = image_utils.blend(
            tf.zeros_like(images), images, value_factors, (0, 255)
        )
        return images

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "value_range": self.value_range,
                "hue_factor": self.hue_factor,
                "saturation_factor": self.saturation_factor,
                "value_factor": self.value_factor,
                "seed": self.seed,
            }
        )
        return config



import numpy as np
import tensorflow as tf
from absl.testing import parameterized

from keras_aug import layers
from keras_aug.datapoints import bounding_box


class ResizeTest(tf.test.TestCase, parameterized.TestCase):
    height, width = 224, 224
    regular_args = {
        "height": 224,
        "width": 224,
        "interpolation": "bilinear",
        "antialias": False,
        "bounding_box_format": "rel_xyxy",
    }
    resize_with_crop_args = {
        "height": 224,
        "width": 224,
        "interpolation": "bilinear",
        "antialias": False,
        "crop_to_aspect_ratio": True,
        "position": "center",
        "padding_value": 0,
        "bounding_box_format": "rel_xyxy",
    }
    resize_with_pad_args = {
        "height": 224,
        "width": 224,
        "interpolation": "bilinear",
        "antialias": False,
        "pad_to_aspect_ratio": True,
        "position": "center",
        "padding_value": 0,
        "bounding_box_format": "rel_xyxy",
    }

    """Resize"""

    def test_augments_image(self):
        input_image_shape = (4, 200, 300, 3)
        image = tf.random.uniform(shape=input_image_shape)
        layer = layers.Resize(**self.regular_args, seed=2023)
        input_image_resized = tf.image.resize(image, [self.height, self.width])

        output = layer(image)

        self.assertAllClose(output, input_image_resized)

    def test_grayscale(self):
        input_image_shape = (4, self.height, self.width, 1)
        image = tf.random.uniform(shape=input_image_shape)
        layer = layers.Resize(**self.regular_args, seed=2023)

        output = layer(image)

        self.assertAllEqual(output.shape, (4, self.height, self.width, 1))

    def test_augment_sparse_segmentation_mask(self):
        num_classes = 8
        input_image_shape = (1, 300, 300, 3)
        mask_shape = (1, 300, 300, 1)
        image = tf.random.uniform(shape=input_image_shape, seed=2023)
        mask = np.random.randint(2, size=mask_shape) * (num_classes - 1)
        inputs = {"images": image, "segmentation_masks": mask}

        # Crop-only to exactly 1/2 of the size
        args = self.regular_args.copy()
        args.update({"height": 150, "width": 150})
        layer = layers.Resize(**args, seed=2023)
        input_mask_resized = tf.image.resize(mask, (150, 150), "nearest")

        output = layer(inputs)

        self.assertAllClose(output["segmentation_masks"], input_mask_resized)

        # Crop to an arbitrary size and make sure we don't do bad interpolation
        args = self.regular_args.copy()
        args.update({"height": 233, "width": 233})
        layer = layers.Resize(**args, seed=2023)

        output = layer(inputs)

        self.assertAllInSet(output["segmentation_masks"], [0, 7])

    def test_augment_bounding_box_single(self):
        image = tf.zeros([20, 20, 3])
        boxes = {
            "boxes": tf.convert_to_tensor([[0, 0, 1, 1]]),
            "classes": tf.convert_to_tensor([0]),
        }
        input = {"images": image, "bounding_boxes": boxes}
        args = self.regular_args.copy()
        args.update({"height": 10, "width": 10})
        layer = layers.Resize(**args, seed=2023)
        expected_output = {
            "boxes": tf.convert_to_tensor([[0, 0, 1, 1]], dtype=tf.float32),
            "classes": tf.convert_to_tensor([0], dtype=tf.float32),
        }

        output = layer(input)
        output["bounding_boxes"] = bounding_box.to_dense(
            output["bounding_boxes"]
        )

        self.assertAllClose(
            expected_output["boxes"], output["bounding_boxes"]["boxes"]
        )
        self.assertAllClose(
            expected_output["classes"], output["bounding_boxes"]["classes"]
        )

    def test_augment_boxes_batched_input(self):
        image = tf.zeros([20, 20, 3])
        boxes = {
            "boxes": tf.convert_to_tensor(
                [
                    [[0, 0, 1, 1], [0, 0, 1, 1]],
                    [[0, 0, 1, 1], [0, 0, 1, 1]],
                ]
            ),
            "classes": tf.convert_to_tensor([[0, 0], [0, 0]]),
        }
        input = {"images": [image, image], "bounding_boxes": boxes}
        args = self.regular_args.copy()
        args.update({"height": 18, "width": 18})
        layer = layers.Resize(**args, seed=2023)
        expected_output = {
            "boxes": tf.convert_to_tensor(
                [
                    [[0, 0, 1, 1], [0, 0, 1, 1]],
                    [[0, 0, 1, 1], [0, 0, 1, 1]],
                ]
            ),
            "classes": tf.convert_to_tensor([[0, 0], [0, 0]]),
        }

        output = layer(input)
        output["bounding_boxes"] = bounding_box.to_dense(
            output["bounding_boxes"]
        )

        self.assertAllClose(
            expected_output["boxes"], output["bounding_boxes"]["boxes"]
        )
        self.assertAllClose(
            expected_output["classes"], output["bounding_boxes"]["classes"]
        )

    def test_augment_boxes_ragged(self):
        image = tf.zeros([2, 20, 20, 3])
        boxes = {
            "boxes": tf.ragged.constant(
                [[[0, 0, 1, 1], [0, 0, 1, 1]], [[0, 0, 1, 1]]], dtype=tf.float32
            ),
            "classes": tf.ragged.constant([[0, 0], [0]]),
        }
        input = {"images": image, "bounding_boxes": boxes}
        args = self.regular_args.copy()
        args.update({"height": 18, "width": 18})
        layer = layers.Resize(**args, seed=2023)
        # the result boxes will still have the entire image in them
        expected_output = {
            "boxes": tf.ragged.constant(
                [[[0, 0, 1, 1], [0, 0, 1, 1]], [[0, 0, 1, 1]]], dtype=tf.float32
            ),
            "classes": tf.ragged.constant([[0, 0], [0]]),
        }
        expected_output = bounding_box.to_dense(expected_output)

        output = layer(input)
        output["bounding_boxes"] = bounding_box.to_dense(
            output["bounding_boxes"]
        )

        self.assertAllClose(
            expected_output["boxes"], output["bounding_boxes"]["boxes"]
        )
        self.assertAllClose(
            expected_output["classes"], output["bounding_boxes"]["classes"]
        )

    def _run_output_shape_test(self, kwargs, height, width):
        kwargs.update({"height": height, "width": width})
        layer = layers.Resize(**kwargs)
        inputs = tf.random.uniform((2, 5, 8, 3))

        outputs = layer(inputs)

        self.assertEqual(outputs.shape, (2, height, width, 3))

    @parameterized.named_parameters(
        ("down_sample_bilinear_2_by_2", {"interpolation": "bilinear"}, 2, 2),
        ("down_sample_bilinear_3_by_2", {"interpolation": "bilinear"}, 3, 2),
        ("down_sample_nearest_2_by_2", {"interpolation": "nearest"}, 2, 2),
        ("down_sample_nearest_3_by_2", {"interpolation": "nearest"}, 3, 2),
        ("down_sample_area_2_by_2", {"interpolation": "area"}, 2, 2),
        ("down_sample_area_3_by_2", {"interpolation": "area"}, 3, 2),
    )
    def test_down_sampling(self, kwargs, height, width):
        self._run_output_shape_test(kwargs, height, width)

    @parameterized.named_parameters(
        ("up_sample_bilinear_10_by_12", {"interpolation": "bilinear"}, 10, 12),
        ("up_sample_bilinear_12_by_12", {"interpolation": "bilinear"}, 12, 12),
        ("up_sample_nearest_10_by_12", {"interpolation": "nearest"}, 10, 12),
        ("up_sample_nearest_12_by_12", {"interpolation": "nearest"}, 12, 12),
        ("up_sample_area_10_by_12", {"interpolation": "area"}, 10, 12),
        ("up_sample_area_12_by_12", {"interpolation": "area"}, 12, 12),
    )
    def test_up_sampling(self, kwargs, expected_height, expected_width):
        self._run_output_shape_test(kwargs, expected_height, expected_width)

    """Resize with crop_to_aspect_ratio"""

    def test_crop_augments_image(self):
        input_image_shape = (4, 200, 300, 3)
        image = tf.random.uniform(shape=input_image_shape)
        layer = layers.Resize(**self.resize_with_crop_args, seed=2023)
        input_image_resized = tf.image.resize(image, [self.height, self.width])

        output = layer(image)

        self.assertNotAllClose(output, input_image_resized)

    def test_crop_augment_sparse_segmentation_mask(self):
        num_classes = 8
        input_image_shape = (1, 300, 300, 3)
        mask_shape = (1, 300, 300, 1)
        image = tf.random.uniform(shape=input_image_shape, seed=2023)
        mask = np.random.randint(2, size=mask_shape) * (num_classes - 1)
        inputs = {"images": image, "segmentation_masks": mask}

        # Crop-only to exactly 1/2 of the size
        args = self.resize_with_crop_args.copy()
        args.update({"height": 150, "width": 150})
        layer = layers.Resize(**args, seed=2023)
        input_mask_resized = tf.image.resize(mask, (150, 150), "nearest")

        output = layer(inputs)

        self.assertAllClose(output["segmentation_masks"], input_mask_resized)

        # Crop to an arbitrary size and make sure we don't do bad interpolation
        args = self.resize_with_crop_args.copy()
        args.update({"height": 233, "width": 233})
        layer = layers.Resize(**args, seed=2023)

        output = layer(inputs)

        self.assertAllInSet(output["segmentation_masks"], [0, 7])

    def test_crop_augment_bounding_box_single(self):
        image = tf.zeros([20, 20, 3])
        boxes = {
            "boxes": tf.convert_to_tensor([[0, 0, 1, 1]]),
            "classes": tf.convert_to_tensor([0]),
        }
        input = {"images": image, "bounding_boxes": boxes}
        args = self.resize_with_crop_args.copy()
        args.update({"height": 10, "width": 10})
        layer = layers.Resize(**args, seed=2023)
        expected_output = {
            "boxes": tf.convert_to_tensor([[0, 0, 1, 1]], dtype=tf.float32),
            "classes": tf.convert_to_tensor([0], dtype=tf.float32),
        }

        output = layer(input)
        output["bounding_boxes"] = bounding_box.to_dense(
            output["bounding_boxes"]
        )

        self.assertAllClose(
            expected_output["boxes"], output["bounding_boxes"]["boxes"]
        )
        self.assertAllClose(
            expected_output["classes"], output["bounding_boxes"]["classes"]
        )

    def test_crop_augment_boxes_batched_input(self):
        image = tf.zeros([20, 20, 3])
        boxes = {
            "boxes": tf.convert_to_tensor(
                [
                    [[0, 0, 1, 1], [0, 0, 1, 1]],
                    [[0, 0, 1, 1], [0, 0, 1, 1]],
                ]
            ),
            "classes": tf.convert_to_tensor([[0, 0], [0, 0]]),
        }
        input = {"images": [image, image], "bounding_boxes": boxes}
        args = self.resize_with_crop_args.copy()
        args.update({"height": 18, "width": 18})
        layer = layers.Resize(**args, seed=2023)
        expected_output = {
            "boxes": tf.convert_to_tensor(
                [
                    [[0, 0, 1, 1], [0, 0, 1, 1]],
                    [[0, 0, 1, 1], [0, 0, 1, 1]],
                ]
            ),
            "classes": tf.convert_to_tensor([[0, 0], [0, 0]]),
        }

        output = layer(input)
        output["bounding_boxes"] = bounding_box.to_dense(
            output["bounding_boxes"]
        )

        self.assertAllClose(
            expected_output["boxes"], output["bounding_boxes"]["boxes"]
        )
        self.assertAllClose(
            expected_output["classes"], output["bounding_boxes"]["classes"]
        )

    def test_crop_augment_boxes_ragged(self):
        image = tf.zeros([2, 20, 20, 3])
        boxes = {
            "boxes": tf.ragged.constant(
                [[[0, 0, 1, 1], [0, 0, 1, 1]], [[0, 0, 1, 1]]], dtype=tf.float32
            ),
            "classes": tf.ragged.constant([[0, 0], [0]]),
        }
        input = {"images": image, "bounding_boxes": boxes}
        args = self.resize_with_crop_args.copy()
        args.update({"height": 18, "width": 18})
        layer = layers.Resize(**args, seed=2023)
        # the result boxes will still have the entire image in them
        expected_output = {
            "boxes": tf.ragged.constant(
                [[[0, 0, 1, 1], [0, 0, 1, 1]], [[0, 0, 1, 1]]], dtype=tf.float32
            ),
            "classes": tf.ragged.constant([[0, 0], [0]]),
        }
        expected_output = bounding_box.to_dense(expected_output)

        output = layer(input)
        output["bounding_boxes"] = bounding_box.to_dense(
            output["bounding_boxes"]
        )

        self.assertAllClose(
            expected_output["boxes"], output["bounding_boxes"]["boxes"]
        )
        self.assertAllClose(
            expected_output["classes"], output["bounding_boxes"]["classes"]
        )

    def test_cropping_center(self):
        inputs = tf.random.uniform((1, 4, 8, 3))
        args = self.resize_with_crop_args.copy()
        args.update({"height": 4, "width": 4, "position": "center"})
        layer = layers.Resize(**args)

        outputs = layer(inputs)

        self.assertEqual(outputs.shape, (1, 4, 4, 3))
        self.assertAllEqual(outputs, inputs[:, :, 2:-2, :])

    """Resize with pad_to_aspect_ratio"""

    def test_pad_augments_image(self):
        input_image_shape = (4, 200, 300, 3)
        image = tf.random.uniform(shape=input_image_shape)
        layer = layers.Resize(**self.resize_with_pad_args, seed=2023)
        input_image_resized = tf.image.resize(image, [self.height, self.width])

        output = layer(image)

        self.assertNotAllClose(output, input_image_resized)

    def test_pad_augment_sparse_segmentation_mask(self):
        num_classes = 8
        input_image_shape = (1, 300, 300, 3)
        mask_shape = (1, 300, 300, 1)
        image = tf.random.uniform(shape=input_image_shape, seed=2023)
        mask = np.random.randint(2, size=mask_shape) * (num_classes - 1)
        inputs = {"images": image, "segmentation_masks": mask}

        # Crop-only to exactly 1/2 of the size
        args = self.resize_with_pad_args.copy()
        args.update({"height": 150, "width": 150})
        layer = layers.Resize(**args, seed=2023)
        input_mask_resized = tf.image.resize(mask, (150, 150), "nearest")

        output = layer(inputs)

        self.assertAllClose(output["segmentation_masks"], input_mask_resized)

        # Crop to an arbitrary size and make sure we don't do bad interpolation
        args = self.regular_args.copy()
        args.update({"height": 233, "width": 233})
        layer = layers.Resize(**args, seed=2023)

        output = layer(inputs)

        self.assertAllInSet(output["segmentation_masks"], [0, 7])

    def test_pad_augment_bounding_box_single(self):
        image = tf.zeros([20, 20, 3])
        boxes = {
            "boxes": tf.convert_to_tensor([[0, 0, 1, 1]]),
            "classes": tf.convert_to_tensor([0]),
        }
        input = {"images": image, "bounding_boxes": boxes}
        args = self.resize_with_pad_args.copy()
        args.update({"height": 10, "width": 10})
        layer = layers.Resize(**args, seed=2023)
        expected_output = {
            "boxes": tf.convert_to_tensor([[0, 0, 1, 1]], dtype=tf.float32),
            "classes": tf.convert_to_tensor([0], dtype=tf.float32),
        }

        output = layer(input)
        output["bounding_boxes"] = bounding_box.to_dense(
            output["bounding_boxes"]
        )

        self.assertAllClose(
            expected_output["boxes"], output["bounding_boxes"]["boxes"]
        )
        self.assertAllClose(
            expected_output["classes"], output["bounding_boxes"]["classes"]
        )

    def test_pad_augment_boxes_batched_input(self):
        image = tf.zeros([20, 20, 3])
        boxes = {
            "boxes": tf.convert_to_tensor(
                [
                    [[0, 0, 1, 1], [0, 0, 1, 1]],
                    [[0, 0, 1, 1], [0, 0, 1, 1]],
                ]
            ),
            "classes": tf.convert_to_tensor([[0, 0], [0, 0]]),
        }
        input = {"images": [image, image], "bounding_boxes": boxes}
        args = self.resize_with_pad_args.copy()
        args.update({"height": 18, "width": 18})
        layer = layers.Resize(**args, seed=2023)
        expected_output = {
            "boxes": tf.convert_to_tensor(
                [
                    [[0, 0, 1, 1], [0, 0, 1, 1]],
                    [[0, 0, 1, 1], [0, 0, 1, 1]],
                ]
            ),
            "classes": tf.convert_to_tensor([[0, 0], [0, 0]]),
        }

        output = layer(input)
        output["bounding_boxes"] = bounding_box.to_dense(
            output["bounding_boxes"]
        )

        self.assertAllClose(
            expected_output["boxes"], output["bounding_boxes"]["boxes"]
        )
        self.assertAllClose(
            expected_output["classes"], output["bounding_boxes"]["classes"]
        )

    def test_pad_augment_boxes_ragged(self):
        image = tf.zeros([2, 20, 20, 3])
        boxes = {
            "boxes": tf.ragged.constant(
                [[[0, 0, 1, 1], [0, 0, 1, 1]], [[0, 0, 1, 1]]], dtype=tf.float32
            ),
            "classes": tf.ragged.constant([[0, 0], [0]]),
        }
        input = {"images": image, "bounding_boxes": boxes}
        args = self.resize_with_pad_args.copy()
        args.update({"height": 18, "width": 18})
        layer = layers.Resize(**args, seed=2023)
        # the result boxes will still have the entire image in them
        expected_output = {
            "boxes": tf.ragged.constant(
                [[[0, 0, 1, 1], [0, 0, 1, 1]], [[0, 0, 1, 1]]], dtype=tf.float32
            ),
            "classes": tf.ragged.constant([[0, 0], [0]]),
        }
        expected_output = bounding_box.to_dense(expected_output)

        output = layer(input)
        output["bounding_boxes"] = bounding_box.to_dense(
            output["bounding_boxes"]
        )

        self.assertAllClose(
            expected_output["boxes"], output["bounding_boxes"]["boxes"]
        )
        self.assertAllClose(
            expected_output["classes"], output["bounding_boxes"]["classes"]
        )

    def test_padding_center(self):
        inputs = tf.ones((1, 4, 8, 3))
        args = self.resize_with_pad_args.copy()
        args.update({"height": 8, "width": 8, "position": "center"})
        layer = layers.Resize(**args)

        outputs = layer(inputs)

        self.assertEqual(outputs.shape, (1, 8, 8, 3))
        self.assertEqual(tf.reduce_mean(outputs[:, 0:2, :, :]), 0.0)
        self.assertEqual(tf.reduce_mean(outputs[:, -2:, :, :]), 0.0)

    def test_padding_top(self):
        inputs = tf.ones((1, 4, 8, 3))
        args = self.resize_with_pad_args.copy()
        args.update({"height": 8, "width": 8, "position": "top_left"})
        layer = layers.Resize(**args)

        outputs = layer(inputs)

        self.assertEqual(outputs.shape, (1, 8, 8, 3))
        self.assertNotEqual(tf.reduce_mean(outputs[:, 0:2, :, :]), 0.0)
        self.assertEqual(tf.reduce_mean(outputs[:, -4:, :, :]), 0.0)

    def test_padding_bottom(self):
        inputs = tf.ones((1, 4, 8, 3))
        args = self.resize_with_pad_args.copy()
        args.update({"height": 8, "width": 8, "position": "bottom_left"})
        layer = layers.Resize(**args)

        outputs = layer(inputs)

        self.assertEqual(outputs.shape, (1, 8, 8, 3))
        self.assertEqual(tf.reduce_mean(outputs[:, 0:4, :, :]), 0.0)
        self.assertNotEqual(tf.reduce_mean(outputs[:, -2:, :, :]), 0.0)

    def test_padding_left(self):
        inputs = tf.ones((1, 8, 4, 3))
        args = self.resize_with_pad_args.copy()
        args.update({"height": 8, "width": 8, "position": "top_left"})
        layer = layers.Resize(**args)

        outputs = layer(inputs)

        self.assertEqual(outputs.shape, (1, 8, 8, 3))
        self.assertNotEqual(tf.reduce_mean(outputs[:, :, 0:2, :]), 0.0)
        self.assertEqual(tf.reduce_mean(outputs[:, :, -2:, :]), 0.0)

    def test_padding_right(self):
        inputs = tf.ones((1, 8, 4, 3))
        args = self.resize_with_pad_args.copy()
        args.update({"height": 8, "width": 8, "position": "top_right"})
        layer = layers.Resize(**args)

        outputs = layer(inputs)

        self.assertEqual(outputs.shape, (1, 8, 8, 3))
        self.assertEqual(tf.reduce_mean(outputs[:, :, 0:4, :]), 0.0)
        self.assertNotEqual(tf.reduce_mean(outputs[:, :, -2:, :]), 0.0)

    def test_dense_segmentation_masks(self):
        images = tf.random.uniform((2, 10, 10, 3))
        segmentation_masks = tf.random.uniform(
            (2, 10, 10, 1), minval=0, maxval=10, dtype=tf.int32
        )
        args = self.resize_with_pad_args.copy()
        args.update({"height": 8, "width": 8, "position": "top_right"})
        layer = layers.Resize(**args)

        result = layer(
            {"images": images, "segmentation_masks": segmentation_masks}
        )

        self.assertTrue(isinstance(result["segmentation_masks"], tf.Tensor))
        self.assertEqual(result["segmentation_masks"].shape[1:3], (8, 8))
        self.assertAllInSet(result["segmentation_masks"], tf.range(0, 10))

    def test_ragged_segmentation_masks(self):
        images = tf.ragged.stack(
            [
                tf.random.uniform((8, 8, 3), dtype=tf.float32),
                tf.random.uniform((16, 8, 3), dtype=tf.float32),
            ]
        )
        segmentation_masks = tf.ragged.stack(
            [
                tf.random.uniform((8, 8, 1), maxval=10, dtype=tf.int32),
                tf.random.uniform((16, 8, 1), maxval=10, dtype=tf.int32),
            ]
        )
        segmentation_masks = tf.cast(segmentation_masks, dtype=tf.float32)
        args = self.resize_with_pad_args.copy()
        args.update({"height": 8, "width": 8, "position": "top_right"})
        layer = layers.Resize(**args)

        result = layer(
            {"images": images, "segmentation_masks": segmentation_masks}
        )

        self.assertTrue(isinstance(result["segmentation_masks"], tf.Tensor))
        self.assertEqual(result["segmentation_masks"].shape[1:3], (8, 8))
        self.assertAllInSet(result["segmentation_masks"], tf.range(0, 10))

    def test_ragged_input_with_graph_mode(self):
        images = tf.ragged.stack(
            [
                tf.random.uniform((8, 8, 3), dtype=tf.float32),
                tf.random.uniform((16, 8, 3), dtype=tf.float32),
            ]
        )
        segmentation_masks = tf.ragged.stack(
            [
                tf.random.uniform((8, 8, 1), maxval=10, dtype=tf.int32),
                tf.random.uniform((16, 8, 1), maxval=10, dtype=tf.int32),
            ]
        )
        args = self.regular_args.copy()
        args.update({"height": 8, "width": 8})
        layer = layers.Resize(**args)

        @tf.function
        def fn(inputs):
            outputs = layer(inputs)
            image_shape = outputs["images"].shape
            segmentation_mask_shape = outputs["segmentation_masks"].shape
            assert image_shape == (2, 8, 8, 3)
            assert segmentation_mask_shape == (2, 8, 8, 1)

        fn({"images": images, "segmentation_masks": segmentation_masks})



"""
Most of these codes come from KerasCV.
"""
import tensorflow as tf

from keras_aug.datapoints import bounding_box


class ToDenseTest(tf.test.TestCase):
    def test_converts_to_dense(self):
        bounding_boxes = {
            "boxes": tf.ragged.constant(
                [[[0, 0, 1, 1]], [[0, 0, 1, 1], [0, 0, 1, 1], [0, 0, 1, 1]]]
            ),
            "classes": tf.ragged.constant([[0], [1, 2, 3]]),
        }

        bounding_boxes = bounding_box.to_dense(bounding_boxes)

        self.assertEqual(bounding_boxes["boxes"].shape, [2, 3, 4])
        self.assertEqual(bounding_boxes["classes"].shape, [2, 3])

    def test_converts_to_dense_with_max_boxes(self):
        bounding_boxes = {
            "boxes": tf.ragged.constant(
                [[[0, 0, 1, 1]], [[0, 0, 1, 1], [0, 0, 1, 1], [0, 0, 1, 1]]]
            ),
            "classes": tf.ragged.constant([[0], [1, 2, 3]]),
        }

        bounding_boxes = bounding_box.to_dense(bounding_boxes, max_boxes=16)

        self.assertEqual(bounding_boxes["boxes"].shape, [2, 16, 4])
        self.assertEqual(bounding_boxes["classes"].shape, [2, 16])



import tensorflow as tf
from tensorflow import keras


def get_rotation_matrix(
    angles, image_height, image_width, to_square=False, name=None
):
    """Returns projective transforms for the given angles.

    Args:
        angles (tf.Tensor): a vector with the angles to rotate each image
            in the batch.
        image_height (tf.Tensor): Height of the images to be transformed.
        image_width (tf.Tensor): Width of the images to be transformed.
        to_square (bool, optional): Whether to append ones to last dimension
            and reshape to ``(batch_size, 3, 3)``. Defaults to ``False``.
        name (str, optional): The name of the op. Defaults to ``None``.

    References:
        - `KerasCV <https://github.com/keras-team/keras-cv>`_
    """
    with keras.backend.name_scope(name or "rotation_matrix"):
        x_offset = (image_width - 1) - (
            tf.cos(angles) * (image_width - 1)
            - tf.sin(angles) * (image_height - 1)
        )
        x_offset /= 2.0
        y_offset = (image_height - 1) - (
            tf.sin(angles) * (image_width - 1)
            + tf.cos(angles) * (image_height - 1)
        )
        y_offset /= 2.0
        num_angles = tf.shape(angles)[0]
        matrix = tf.concat(
            [
                tf.cos(angles),
                -tf.sin(angles),
                x_offset,
                tf.sin(angles),
                tf.cos(angles),
                y_offset,
                tf.zeros((num_angles, 2), angles.dtype),
            ],
            axis=1,
        )
        if to_square:
            matrix = tf.concat(
                [matrix, tf.ones((num_angles, 1), angles.dtype)], axis=1
            )
            matrix = tf.reshape(matrix, (num_angles, 3, 3))
        return matrix


def get_translation_matrix(
    translations, image_height, image_width, to_square=False, name=None
):
    """Returns projective transforms for the given translations.

    Args:
        translations (tf.Tensor): A matrix of 2-element lists representing
            ``[dx, dy]`` to translate for a batch of images.
        image_height (tf.Tensor): Height of the images to be transformed.
        image_width (tf.Tensor): Width of the images to be transformed.
        to_square (bool, optional): Whether to append ones to last dimension
            and reshape to ``(batch_size, 3, 3)``. Defaults to ``False``.
        name (str, optional): The name of the op. Defaults to ``None``.

    References:
        - `KerasCV <https://github.com/keras-team/keras-cv>`_
    """
    with keras.backend.name_scope(name or "translation_matrix"):
        num_translations = tf.shape(translations)[0]
        # The translation matrix looks like:
        #     [[1 0 -dx]
        #      [0 1 -dy]
        #      [0 0 1]]
        # where the last entry is implicit.
        # Translation matrices are always float32.
        matrix = tf.concat(
            values=[
                tf.ones((num_translations, 1), translations.dtype),
                tf.zeros((num_translations, 1), translations.dtype),
                -translations[:, 0, tf.newaxis] * image_width,
                tf.zeros((num_translations, 1), translations.dtype),
                tf.ones((num_translations, 1), translations.dtype),
                -translations[:, 1, tf.newaxis] * image_height,
                tf.zeros((num_translations, 2), translations.dtype),
            ],
            axis=1,
        )
        if to_square:
            matrix = tf.concat(
                [matrix, tf.ones((num_translations, 1), translations.dtype)],
                axis=1,
            )
            matrix = tf.reshape(matrix, (num_translations, 3, 3))
        return matrix


def get_zoom_matrix(
    zooms, image_height, image_width, to_square=False, name=None
):
    """Returns projective transforms for the given zooms.

    Args:
        zooms (tf.Tensor): A matrix of 2-element lists representing
            ``[zx, zy]`` to zoom for a batch of images.
        image_height (tf.Tensor): Height of the images to be transformed.
        image_width (tf.Tensor): Width of the images to be transformed.
        to_square (bool, optional): Whether to append ones to last dimension
            and reshape to ``(batch_size, 3, 3)``. Defaults to ``False``.
        name (str, optional): The name of the op. Defaults to ``None``.

    References:
        - `KerasCV <https://github.com/keras-team/keras-cv>`_
    """
    with keras.backend.name_scope(name or "zoom_matrix"):
        num_zooms = tf.shape(zooms)[0]
        # The zoom matrix looks like:
        #     [[zx 0 0]
        #      [0 zy 0]
        #      [0 0 1]]
        # where the last entry is implicit.
        # Zoom matrices are always float32.
        x_offset = ((image_width - 1.0) / 2.0) * (1.0 - zooms[:, 0, tf.newaxis])
        y_offset = ((image_height - 1.0) / 2.0) * (
            1.0 - zooms[:, 1, tf.newaxis]
        )
        matrix = tf.concat(
            values=[
                zooms[:, 0, tf.newaxis],
                tf.zeros((num_zooms, 1), zooms.dtype),
                x_offset,
                tf.zeros((num_zooms, 1), zooms.dtype),
                zooms[:, 1, tf.newaxis],
                y_offset,
                tf.zeros((num_zooms, 2), zooms.dtype),
            ],
            axis=1,
        )
        if to_square:
            matrix = tf.concat(
                [matrix, tf.ones((num_zooms, 1), zooms.dtype)], axis=1
            )
            matrix = tf.reshape(matrix, (num_zooms, 3, 3))
        return matrix


def get_shear_matrix(shears, to_square=False, name=None):
    """Returns projective transforms for the given shears.

    Args:
        shears (tf.Tensor): A matrix of 2-element lists representing `[sx, sy]`
            to shear for a batch of images.
        to_square (bool, optional): Whether to append ones to last dimension
            and reshape to ``(batch_size, 3, 3)``. Defaults to ``False``.
        name (str, optional): The name of the op. Defaults to ``None``.

    References:
        - `KerasCV <https://github.com/keras-team/keras-cv>`_
    """
    with keras.backend.name_scope(name or "shear_matrix"):
        num_shears = tf.shape(shears)[0]
        # The transform matrix looks like:
        # (1, x, 0)
        # (y, 1, 0)
        # (0, 0, 1)
        # where the last entry is implicit.
        matrix = tf.concat(
            values=[
                tf.ones((num_shears, 1), shears.dtype),
                shears[:, 0, tf.newaxis],
                tf.zeros((num_shears, 1), shears.dtype),
                shears[:, 1, tf.newaxis],
                tf.ones((num_shears, 1), shears.dtype),
                tf.zeros((num_shears, 3), shears.dtype),
            ],
            axis=1,
        )
        if to_square:
            matrix = tf.concat(
                [matrix, tf.ones((num_shears, 1), shears.dtype)], axis=1
            )
            matrix = tf.reshape(matrix, (num_shears, 3, 3))
        return matrix



import tensorflow as tf
from tensorflow import keras

from keras_aug.datapoints import image as image_utils
from keras_aug.layers.base.vectorized_base_random_layer import (
    VectorizedBaseRandomLayer,
)
from keras_aug.utils import augmentation as augmentation_utils


@keras.utils.register_keras_serializable(package="keras_aug")
class RandomSharpness(VectorizedBaseRandomLayer):
    """Randomly performs the sharpness operation on given images.

    The sharpness operation first performs a blur operation, then blends between
    the original image and the blurred image. This operation makes the edges of
    an image less sharp than they were in the original image.

    Args:
        value_range (Sequence[int|float]): The range of values the incoming
            images will have. This is typically either ``[0, 1]`` or
            ``[0, 255]`` depending on how your preprocessing pipeline is set up.
        factor (float|Sequence[float]|keras_aug.FactorSampler): The range of the
            sharpness factor. When represented as a single float, the factor
            will be picked between ``[1.0 - lower, 1.0 + upper]``. ``1.0`` will
            give the original image. ``0.0`` makes the image to be blurred.
            ``2.0`` will enhance the sharpness by a factor of 2.
        seed (int|float, optional): The random seed. Defaults to ``None``.

    References:
        - `PIL/ImageEnhance <https://pillow.readthedocs.io/en/stable/reference/ImageEnhance.html>`_
        - `KerasCV <https://github.com/keras-team/keras-cv>`_
        - `Tensorflow Model augment <https://github.com/tensorflow/models/blob/v2.12.0/official/vision/ops/augment.py>`_
    """  # noqa: E501

    def __init__(
        self,
        value_range,
        factor,
        seed=None,
        **kwargs,
    ):
        super().__init__(seed=seed, **kwargs)
        self.value_range = value_range
        self.factor = augmentation_utils.parse_factor(
            factor, max_value=None, center_value=1, seed=seed
        )
        self.seed = seed

    def get_random_transformation_batch(self, batch_size, **kwargs):
        return self.factor(
            shape=(batch_size, 1, 1, 1), dtype=self.compute_dtype
        )

    def augment_ragged_image(self, image, transformation, **kwargs):
        images = tf.expand_dims(image, axis=0)
        new_transformation = tf.expand_dims(transformation, axis=0)
        output = self.augment_images(images, new_transformation)
        return tf.squeeze(output, axis=0)

    def augment_images(self, images, transformations, **kwargs):
        images = image_utils.transform_value_range(
            images,
            original_range=self.value_range,
            target_range=(0, 255),
            dtype=self.compute_dtype,
        )
        original_images = images
        # [1 1 1]
        # [1 5 1]
        # [1 1 1]
        # all divided by 13 is the default 3x3 gaussian smoothing kernel.
        # Correlating or Convolving with this filter is equivalent to performing
        # a gaussian blur.
        kernel = (
            tf.constant(
                [[1, 1, 1], [1, 5, 1], [1, 1, 1]],
                dtype=self.compute_dtype,
                shape=[3, 3, 1, 1],
            )
            / 13.0
        )
        # Tile across channel dimension.
        channels = tf.shape(images)[-1]
        kernel = tf.tile(kernel, [1, 1, channels, 1])
        strides = [1, 1, 1, 1]
        smoothed_image = tf.nn.depthwise_conv2d(
            images, kernel, strides, padding="VALID", dilations=[1, 1]
        )
        smoothed_image = tf.clip_by_value(smoothed_image, 0.0, 255.0)

        # For the borders of the resulting image, fill in the values of the
        # original image.
        mask = tf.ones_like(smoothed_image)
        mask = tf.pad(mask, [[0, 0], [1, 1], [1, 1], [0, 0]])
        smoothed_image = tf.pad(
            smoothed_image, [[0, 0], [1, 1], [1, 1], [0, 0]]
        )

        images = tf.where(tf.equal(mask, 1), smoothed_image, original_images)
        # Blend the final result.
        images = image_utils.blend(
            images, original_images, transformations, (0, 255)
        )
        images = image_utils.transform_value_range(
            images,
            original_range=(0, 255),
            target_range=self.value_range,
            dtype=self.compute_dtype,
        )
        return images

    def augment_labels(self, labels, transformations, **kwargs):
        return labels

    def augment_bounding_boxes(self, bounding_boxes, transformations, **kwargs):
        return bounding_boxes

    def augment_segmentation_masks(
        self, segmentation_masks, transformations, **kwargs
    ):
        return segmentation_masks

    def augment_keypoints(self, keypoints, transformations, **kwargs):
        return keypoints

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "value_range": self.value_range,
                "factor": self.factor,
                "seed": self.seed,
            }
        )
        return config



import tensorflow as tf
from tensorflow import keras

from keras_aug.datapoints import bounding_box
from keras_aug.layers.base.vectorized_base_random_layer import (
    VectorizedBaseRandomLayer,
)
from keras_aug.utils import augmentation as augmentation_utils
from keras_aug.utils.augmentation import BATCHED
from keras_aug.utils.augmentation import BOUNDING_BOXES
from keras_aug.utils.augmentation import IMAGES
from keras_aug.utils.augmentation import LABELS
from keras_aug.utils.augmentation import SEGMENTATION_MASKS


@keras.utils.register_keras_serializable(package="keras_aug")
class Mosaic(VectorizedBaseRandomLayer):
    """The Mosaic data augmentation technique used by YOLO series.

    The Mosaic data augmentation first takes 4 images from the batch and makes a
    grid. After that based on the offset, a crop is taken to form the mosaic
    image. Labels are in the same ratio as the area of their images in the
    output image. Bounding boxes are translated according to the position of the
    4 images.

    Args:
        height (int): The height of result image.
        width (int): The width of result image.
        offset (float|Sequence[float]|keras_aug.FactorSampler): The offset
            of the mosaic center from the top-left corner of the mosaic. If a
            tuple is used, the x and y coordinates of the mosaic center are
            sampled between the two values for every image augmented. When
            represented as a single float, the values will be picked between
            ``[0.5 - offset, 0.5 + offset]``. Defaults to ``(0.25, 0.75)``.
        fill_value (int|float, optional): The value to be filled outside the
            boundaries when ``fill_mode="constant"``. Defaults to ``0``.
        bounding_box_format (str, optional): The format of bounding
            boxes of input dataset. Refer
            https://github.com/james77777778/keras-aug/blob/main/keras_aug/datapoints/bounding_box/converter.py
            for more details on supported bounding box formats.
        seed (int|float, optional): The random seed. Defaults to ``None``.

    References:
        - `YOLOV4 <https://arxiv.org/abs/2004.10934>`_
        - `YOLOX Official Repo <https://github.com/Megvii-BaseDetection/YOLOX>`_
        - `ultralytics/ultralytics <https://github.com/ultralytics/ultralytics>`_
        - `KerasCV <https://github.com/keras-team/keras-cv>`_
    """  # noqa: E501

    def __init__(
        self,
        height,
        width,
        offset=(0.25, 0.75),
        fill_value=0,
        bounding_box_format=None,
        seed=None,
        **kwargs,
    ):
        super().__init__(seed=seed, **kwargs)
        single_image_max_size = max((height, width)) // 2

        self.height = height
        self.width = width
        self.offset = offset
        self.fill_value = fill_value
        self.bounding_box_format = bounding_box_format
        self.seed = seed

        self.center_sampler = augmentation_utils.parse_factor(
            offset,
            min_value=0.0,
            max_value=1.0,
            center_value=0.5,
            seed=seed,
        )
        self.single_image_max_size = single_image_max_size  # for padding

        # set force_no_unwrap_ragged_image_call=True because Mosaic needs
        # to process images in batch.
        # set force_output_dense_images=True because the output images must
        # have same shape (B, height, width, C)
        self.force_no_unwrap_ragged_image_call = True
        self.force_output_dense_images = True
        self.force_output_dense_segmentation_masks = True

    def get_random_transformation_batch(self, batch_size, **kwargs):
        # pick 3 indices for every batch to create the mosaic output with.
        permutation_order = self._random_generator.random_uniform(
            (batch_size, 3),
            minval=0,
            maxval=batch_size,
            dtype=tf.int32,
        )
        # concatenate the batches with permutation order to get all 4 images of
        # the mosaic
        permutation_order = tf.concat(
            [tf.expand_dims(tf.range(batch_size), axis=-1), permutation_order],
            axis=-1,
        )

        mosaic_centers_x = self.center_sampler(shape=(batch_size,))
        mosaic_centers_y = self.center_sampler(shape=(batch_size,))
        mosaic_centers = tf.stack((mosaic_centers_x, mosaic_centers_y), axis=-1)

        return {
            "permutation_order": permutation_order,
            "mosaic_centers": mosaic_centers,
        }

    def augment_images(self, images, transformations, **kwargs):
        ori_shape = images.shape
        permutation_order = transformations["permutation_order"]
        mosaic_images = tf.gather(images, permutation_order)
        inputs_for_pad_and_mosaic_single_image = {
            "transformations": transformations,
            IMAGES: mosaic_images,
        }
        images = tf.map_fn(
            self.get_mosaic_single_image,
            inputs_for_pad_and_mosaic_single_image,
            fn_output_signature=self.compute_dtype,
        )
        images = tf.ensure_shape(
            images,
            shape=(ori_shape[0], self.height, self.width, ori_shape[-1]),
        )
        return images

    def augment_labels(self, labels, transformations, images=None, **kwargs):
        # cast to float32 to avoid numerical issue
        is_ragged_labels = isinstance(labels, tf.RaggedTensor)
        if is_ragged_labels:
            labels = labels.to_tensor()
        labels = tf.cast(labels, dtype=tf.float32)
        heights, widths = augmentation_utils.get_images_shape(
            images, dtype=tf.float32
        )
        # updates labels for one output mosaic
        permutation_order = transformations["permutation_order"]
        labels_for_mosaic = tf.gather(labels, permutation_order)

        mosaic_centers = transformations["mosaic_centers"]
        center_xs = tf.expand_dims(mosaic_centers[..., 0], axis=-1) * widths
        center_ys = tf.expand_dims(mosaic_centers[..., 1], axis=-1) * heights

        areas = heights * widths

        # labels are in the same ratio as the area of the images
        top_left_ratio = (center_xs * center_ys) / areas
        top_right_ratio = ((widths - center_xs) * center_ys) / areas
        bottom_left_ratio = (center_xs * (heights - center_ys)) / areas
        bottom_right_ratio = (
            (widths - center_xs) * (heights - center_ys)
        ) / areas
        labels = (
            labels_for_mosaic[:, 0] * top_left_ratio
            + labels_for_mosaic[:, 1] * top_right_ratio
            + labels_for_mosaic[:, 2] * bottom_left_ratio
            + labels_for_mosaic[:, 3] * bottom_right_ratio
        )
        if is_ragged_labels:
            labels = tf.RaggedTensor.from_tensor(labels)
        return tf.cast(labels, dtype=self.compute_dtype)

    def augment_bounding_boxes(
        self, bounding_boxes, transformations, raw_images=None, **kwargs
    ):
        batch_size = tf.shape(raw_images)[0]
        heights, widths = augmentation_utils.get_images_shape(
            raw_images, dtype=tf.float32
        )
        bounding_boxes = bounding_box.to_dense(bounding_boxes)
        bounding_boxes = bounding_box.convert_format(
            bounding_boxes,
            source=self.bounding_box_format,
            target="xyxy",
            images=raw_images,
            dtype=tf.float32,
        )
        boxes, classes = bounding_boxes["boxes"], bounding_boxes["classes"]

        # values to translate the boxes by in the mosaic image
        mosaic_centers = transformations["mosaic_centers"]
        mosaic_centers_x = tf.expand_dims(
            mosaic_centers[..., 0] * self.width, axis=-1
        )
        mosaic_centers_y = tf.expand_dims(
            mosaic_centers[..., 1] * self.height, axis=-1
        )

        # updates bounding_boxes for one output mosaic
        permutation_order = transformations["permutation_order"]
        heights_for_mosaic = tf.gather(heights, permutation_order)
        widths_for_mosaic = tf.gather(widths, permutation_order)
        classes_for_mosaic = tf.gather(classes, permutation_order)
        boxes_for_mosaic = tf.gather(boxes, permutation_order)

        # translate_xs/translate_ys 3D:
        # (batch_size, mosaic_index, 1)
        translate_xs = tf.stack(
            (
                mosaic_centers_x - widths_for_mosaic[:, 0, ...],
                mosaic_centers_x,
                mosaic_centers_x - widths_for_mosaic[:, 2, ...],
                mosaic_centers_x,
            ),
            axis=1,
        )
        translate_ys = tf.stack(
            (
                mosaic_centers_y - heights_for_mosaic[:, 0, ...],
                mosaic_centers_y - heights_for_mosaic[:, 1, ...],
                mosaic_centers_y,
                mosaic_centers_y,
            ),
            axis=1,
        )
        # translate_xs/translate_ys 4D:
        # (batch_size, mosaic_index, 1, 4)
        # 4 means translates for xyxy
        translates = tf.stack(
            (translate_xs, translate_ys, translate_xs, translate_ys), axis=-1
        )
        # boxes_for_mosaic 4D:
        # (batch_size, mosaic_index, num_boxes, coordinates)
        boxes_for_mosaic += translates

        boxes_for_mosaic = tf.reshape(boxes_for_mosaic, [batch_size, -1, 4])
        classes_for_mosaic = tf.reshape(classes_for_mosaic, [batch_size, -1])
        boxes_for_mosaic = {
            "boxes": boxes_for_mosaic,
            "classes": classes_for_mosaic,
        }
        boxes_for_mosaic = bounding_box.clip_to_image(
            boxes_for_mosaic,
            bounding_box_format="xyxy",
            image_shape=(self.height, self.width, None),
        )
        boxes_for_mosaic = bounding_box.convert_format(
            boxes_for_mosaic,
            source="xyxy",
            target=self.bounding_box_format,
            image_shape=(self.height, self.width, None),
            dtype=self.compute_dtype,
        )
        return boxes_for_mosaic

    def augment_segmentation_masks(
        self, segmentation_masks, transformations, **kwargs
    ):
        ori_shape = segmentation_masks.shape
        permutation_order = transformations["permutation_order"]
        mosaic_segmentation_masks = tf.gather(
            segmentation_masks, permutation_order
        )
        inputs_for_pad_and_mosaic_single_image = {
            "transformations": transformations,
            IMAGES: mosaic_segmentation_masks,
        }
        mosaic_segmentation_masks = tf.map_fn(
            self.get_mosaic_single_image,
            inputs_for_pad_and_mosaic_single_image,
            fn_output_signature=self.compute_dtype,
        )
        mosaic_segmentation_masks = tf.ensure_shape(
            mosaic_segmentation_masks,
            shape=(ori_shape[0], self.height, self.width, ori_shape[-1]),
        )
        return mosaic_segmentation_masks

    def _batch_augment(self, inputs):
        self._validate_inputs(inputs)
        return super()._batch_augment(inputs)

    def call(self, inputs):
        _, metadata = self._format_inputs(inputs)
        if metadata[BATCHED] is not True:
            raise ValueError(
                "Mosaic received a single image to `call`. The "
                "layer relies on combining multiple examples, and as such "
                "will not behave as expected. Please call the layer with 4 "
                "or more samples."
            )
        return super().call(inputs=inputs)

    def _validate_inputs(self, inputs):
        images = inputs.get(IMAGES, None)
        labels = inputs.get(LABELS, None)
        bounding_boxes = inputs.get(BOUNDING_BOXES, None)
        segmentation_masks = inputs.get(SEGMENTATION_MASKS, None)
        if images is None or (
            labels is None
            and bounding_boxes is None
            and segmentation_masks is None
        ):
            raise ValueError(
                "Mosaic expects inputs in a dictionary with format "
                '{"images": images, "labels": labels}. or '
                '{"images": images, "bounding_boxes": bounding_boxes} or '
                '{"images": images, "segmentation_masks": segmentation_masks}'
                f"Got: inputs = {inputs}"
            )
        if bounding_boxes is not None and self.bounding_box_format is None:
            raise ValueError(
                "Mosaic received bounding boxes but no "
                "bounding_box_format. Please pass a bounding_box_format from "
                "the supported list."
            )

    def pad_and_crop_image_patch_single(
        self, image, height, width, target_height, target_width, mosaic_index
    ):
        if mosaic_index not in (0, 1, 2, 3):
            ValueError(
                "mosaic_index must be in the range of `[0, 3]`. "
                f"Received mosaic_index: {mosaic_index}"
            )
        padding_row = target_width - width
        padding_col = target_height - height
        # pad
        if mosaic_index == 0:
            paddings = [[padding_col, 0], [padding_row, 0], [0, 0]]
        elif mosaic_index == 1:
            paddings = [[padding_col, 0], [0, padding_row], [0, 0]]
        elif mosaic_index == 2:
            paddings = [[0, padding_col], [padding_row, 0], [0, 0]]
        elif mosaic_index == 3:
            paddings = [[0, padding_col], [0, padding_row], [0, 0]]
        paddings = tf.convert_to_tensor(paddings, dtype=tf.int32)
        paddings = tf.where(paddings > 0, paddings, tf.zeros_like(paddings))
        image = tf.pad(
            image, paddings=paddings, constant_values=self.fill_value
        )
        # crop
        offsets = tf.convert_to_tensor(
            [padding_col, padding_row], dtype=tf.int32
        )
        offsets = tf.where(
            offsets < 0, tf.negative(offsets), tf.zeros_like(offsets)
        )
        if mosaic_index == 0:
            offset_height = offsets[0]
            offset_width = offsets[1]
        elif mosaic_index == 1:
            offset_height = offsets[0]
            offset_width = 0
        elif mosaic_index == 2:
            offset_height = 0
            offset_width = offsets[1]
        elif mosaic_index == 3:
            offset_height = 0
            offset_width = 0
        image = tf.image.crop_to_bounding_box(
            image,
            offset_height=offset_height,
            offset_width=offset_width,
            target_height=target_height,
            target_width=target_width,
        )
        return image

    def get_mosaic_single_image(self, inputs):
        mosaic_images = inputs.get(IMAGES, None)
        transformation = inputs.get("transformations")
        mosaic_centers = transformation["mosaic_centers"]
        # must be tf.int32 for following tf.image.crop_to_bounding_box
        mosaic_centers_x = tf.cast(
            tf.round(mosaic_centers[0] * self.width), dtype=tf.int32
        )
        mosaic_centers_y = tf.cast(
            tf.round(mosaic_centers[1] * self.height), dtype=tf.int32
        )
        heights, widths = augmentation_utils.get_images_shape(
            mosaic_images, dtype=tf.int32
        )

        # get images
        top_left_image = mosaic_images[0]
        top_right_image = mosaic_images[1]
        bottom_left_image = mosaic_images[2]
        bottom_right_image = mosaic_images[3]
        if isinstance(mosaic_images, tf.RaggedTensor):
            top_left_image = top_left_image.to_tensor()
            top_right_image = top_right_image.to_tensor()
            bottom_left_image = bottom_left_image.to_tensor()
            bottom_right_image = bottom_right_image.to_tensor()

        # top_left
        top_left_image = self.pad_and_crop_image_patch_single(
            top_left_image,
            heights[0][0],
            widths[0][0],
            mosaic_centers_y,
            mosaic_centers_x,
            mosaic_index=0,
        )
        # top_right
        top_right_image = self.pad_and_crop_image_patch_single(
            top_right_image,
            heights[1][0],
            widths[1][0],
            mosaic_centers_y,
            self.width - mosaic_centers_x,
            mosaic_index=1,
        )
        # bottom_left
        bottom_left_image = self.pad_and_crop_image_patch_single(
            bottom_left_image,
            heights[2][0],
            widths[2][0],
            self.height - mosaic_centers_y,
            mosaic_centers_x,
            mosaic_index=2,
        )
        # bottom_right
        bottom_right_image = self.pad_and_crop_image_patch_single(
            bottom_right_image,
            heights[3][0],
            widths[3][0],
            self.height - mosaic_centers_y,
            self.width - mosaic_centers_x,
            mosaic_index=3,
        )

        # concat patches
        tops = tf.concat([top_left_image, top_right_image], axis=1)
        bottoms = tf.concat([bottom_left_image, bottom_right_image], axis=1)
        outputs = tf.concat([tops, bottoms], axis=0)
        return tf.cast(outputs, dtype=self.compute_dtype)

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "height": self.height,
                "width": self.width,
                "offset": self.offset,
                "fill_value": self.fill_value,
                "bounding_box_format": self.bounding_box_format,
                "seed": self.seed,
            }
        )
        return config



import tensorflow as tf
from tensorflow import keras

from keras_aug.datapoints import bounding_box
from keras_aug.layers.base.vectorized_base_random_layer import (
    VectorizedBaseRandomLayer,
)
from keras_aug.utils import augmentation as augmentation_utils


@keras.utils.register_keras_serializable(package="keras_aug")
class CenterCrop(VectorizedBaseRandomLayer):
    """Center crops the images.

    CenterCrop crops the central portion of the images to a specified
    ``(height, width)``. If an image is smaller than the target size, it will be
    padded and then cropped.

    Args:
        height (int): The height of result image.
        width (int): The width of result image.
        padding_value (int|float, optional): The padding value.
            Defaults to ``0``.
        bounding_box_format (str, optional): The format of bounding
            boxes of input dataset. Refer
            https://github.com/james77777778/keras-aug/blob/main/keras_aug/datapoints/bounding_box/converter.py
            for more details on supported bounding box formats.
        bounding_box_min_area_ratio (float, optional): The threshold to
            apply sanitize_bounding_boxes. Defaults to ``None``.
        bounding_box_max_aspect_ratio (float, optional): The threshold to
            apply sanitize_bounding_boxes. Defaults to ``None``.
        seed (int|float, optional): The random seed. Defaults to ``None``.
    """  # noqa: E501

    def __init__(
        self,
        height,
        width,
        padding_value=0,
        bounding_box_format=None,
        bounding_box_min_area_ratio=None,
        bounding_box_max_aspect_ratio=None,
        seed=None,
        **kwargs,
    ):
        super().__init__(seed=seed, **kwargs)
        self.height = height
        self.width = width
        self.position = augmentation_utils.get_padding_position("center")
        self.padding_value = padding_value
        self.bounding_box_format = bounding_box_format
        self.bounding_box_min_area_ratio = bounding_box_min_area_ratio
        self.bounding_box_max_aspect_ratio = bounding_box_max_aspect_ratio
        self.seed = seed

        # set force_output_dense_images=True because the output images must
        # have same shape (B, height, width, C)
        self.force_output_dense_images = True

    def get_random_transformation_batch(
        self, batch_size, images=None, **kwargs
    ):
        heights, widths = augmentation_utils.get_images_shape(images)

        tops = tf.where(
            heights < self.height,
            tf.cast((self.height - heights) / 2, heights.dtype),
            tf.zeros_like(heights, dtype=heights.dtype),
        )
        bottoms = tf.where(
            heights < self.height,
            self.height - heights - tops,
            tf.zeros_like(heights, dtype=heights.dtype),
        )
        lefts = tf.where(
            widths < self.width,
            tf.cast((self.width - widths) / 2, widths.dtype),
            tf.zeros_like(widths, dtype=widths.dtype),
        )
        rights = tf.where(
            widths < self.width,
            self.width - widths - lefts,
            tf.zeros_like(widths, dtype=widths.dtype),
        )

        (tops, bottoms, lefts, rights) = augmentation_utils.get_position_params(
            tops, bottoms, lefts, rights, self.position, self._random_generator
        )

        return {
            "pad_tops": tops,
            "pad_bottoms": bottoms,
            "pad_lefts": lefts,
            "pad_rights": rights,
        }

    def compute_ragged_image_signature(self, images):
        return tf.RaggedTensorSpec(
            shape=(self.height, self.width, images.shape[-1]),
            ragged_rank=1,
            dtype=self.compute_dtype,
        )

    def augment_ragged_image(self, image, transformation, **kwargs):
        image = tf.expand_dims(image, axis=0)
        transformation = augmentation_utils.expand_dict_dims(
            transformation, axis=0
        )
        image = self.augment_images(
            images=image, transformations=transformation, **kwargs
        )
        return tf.squeeze(image, axis=0)

    def augment_images(self, images, transformations, **kwargs):
        ori_height = tf.shape(images)[augmentation_utils.H_AXIS]
        ori_width = tf.shape(images)[augmentation_utils.W_AXIS]

        pad_top = transformations["pad_tops"][0][0]
        pad_bottom = transformations["pad_bottoms"][0][0]
        pad_left = transformations["pad_lefts"][0][0]
        pad_right = transformations["pad_rights"][0][0]
        paddings = tf.stack(
            (
                tf.zeros(shape=(2,), dtype=pad_top.dtype),
                tf.stack((pad_top, pad_bottom)),
                tf.stack((pad_left, pad_right)),
                tf.zeros(shape=(2,), dtype=pad_top.dtype),
            )
        )
        images = tf.pad(
            images, paddings=paddings, constant_values=self.padding_value
        )

        # center crop
        offset_height = (ori_height + pad_top + pad_bottom - self.height) // 2
        offset_width = (ori_width + pad_left + pad_right - self.width) // 2
        images = tf.image.crop_to_bounding_box(
            images,
            offset_height,
            offset_width,
            self.height,
            self.width,
        )
        images = tf.ensure_shape(
            images, shape=(None, self.height, self.width, None)
        )
        return tf.cast(images, dtype=self.compute_dtype)

    def augment_labels(self, labels, transformations, **kwargs):
        return labels

    def augment_bounding_boxes(
        self,
        bounding_boxes,
        transformations,
        images=None,
        raw_images=None,
        **kwargs,
    ):
        if self.bounding_box_format is None:
            raise ValueError(
                "`CenterCrop()` was called with bounding boxes,"
                "but no `bounding_box_format` was specified in the constructor."
                "Please specify a bounding box format in the constructor. i.e."
                "`CenterCrop(..., bounding_box_format='xyxy')`"
            )
        bounding_boxes = bounding_box.to_dense(bounding_boxes)
        bounding_boxes = bounding_box.convert_format(
            bounding_boxes,
            source=self.bounding_box_format,
            target="xyxy",
            images=raw_images,
        )
        original_bounding_boxes = bounding_boxes.copy()

        x1s, y1s, x2s, y2s = tf.split(bounding_boxes["boxes"], 4, axis=-1)
        pad_tops = tf.cast(transformations["pad_tops"], dtype=tf.float32)
        pad_lefts = tf.cast(transformations["pad_lefts"], dtype=tf.float32)
        pad_bottoms = tf.cast(transformations["pad_bottoms"], dtype=tf.float32)
        pad_rights = tf.cast(transformations["pad_rights"], dtype=tf.float32)
        heights, widths = augmentation_utils.get_images_shape(
            raw_images, dtype=tf.float32
        )
        offset_heights = (heights + pad_tops + pad_bottoms - self.height) // 2
        offset_widths = (widths + pad_lefts + pad_rights - self.width) // 2

        x1s += tf.expand_dims(pad_lefts - offset_widths, axis=1)
        y1s += tf.expand_dims(pad_tops - offset_heights, axis=1)
        x2s += tf.expand_dims(pad_lefts - offset_widths, axis=1)
        y2s += tf.expand_dims(pad_tops - offset_heights, axis=1)
        outputs = tf.concat([x1s, y1s, x2s, y2s], axis=-1)

        bounding_boxes = bounding_boxes.copy()
        bounding_boxes["boxes"] = outputs
        bounding_boxes = bounding_box.clip_to_image(
            bounding_boxes,
            bounding_box_format="xyxy",
            images=images,
        )
        bounding_boxes = bounding_box.sanitize_bounding_boxes(
            bounding_boxes,
            min_area_ratio=self.bounding_box_min_area_ratio,
            max_aspect_ratio=self.bounding_box_max_aspect_ratio,
            bounding_box_format="xyxy",
            reference_bounding_boxes=original_bounding_boxes,
            images=images,
            reference_images=raw_images,
        )
        bounding_boxes = bounding_box.convert_format(
            bounding_boxes,
            source="xyxy",
            target=self.bounding_box_format,
            dtype=self.compute_dtype,
            images=images,
        )
        return bounding_boxes

    def compute_ragged_segmentation_mask_signature(self, segmentation_masks):
        return tf.RaggedTensorSpec(
            shape=(self.height, self.width, segmentation_masks.shape[-1]),
            ragged_rank=1,
            dtype=self.compute_dtype,
        )

    def augment_ragged_segmentation_mask(
        self, segmentation_mask, transformation, **kwargs
    ):
        segmentation_mask = tf.expand_dims(segmentation_mask, axis=0)
        transformation = augmentation_utils.expand_dict_dims(
            transformation, axis=0
        )
        segmentation_mask = self.augment_segmentation_masks(
            segmentation_masks=segmentation_mask,
            transformations=transformation,
            **kwargs,
        )
        return tf.squeeze(segmentation_mask, axis=0)

    def augment_segmentation_masks(
        self, segmentation_masks, transformations, **kwargs
    ):
        ori_height = tf.shape(segmentation_masks)[augmentation_utils.H_AXIS]
        ori_width = tf.shape(segmentation_masks)[augmentation_utils.W_AXIS]
        pad_top = transformations["pad_tops"][0][0]
        pad_bottom = transformations["pad_bottoms"][0][0]
        pad_left = transformations["pad_lefts"][0][0]
        pad_right = transformations["pad_rights"][0][0]
        paddings = tf.stack(
            (
                tf.zeros(shape=(2,), dtype=pad_top.dtype),
                tf.stack((pad_top, pad_bottom)),
                tf.stack((pad_left, pad_right)),
                tf.zeros(shape=(2,), dtype=pad_top.dtype),
            )
        )
        segmentation_masks = tf.pad(
            segmentation_masks, paddings=paddings, constant_values=0
        )
        # center crop
        offset_height = (ori_height + pad_top + pad_bottom - self.height) // 2
        offset_width = (ori_width + pad_left + pad_right - self.width) // 2
        segmentation_masks = tf.image.crop_to_bounding_box(
            segmentation_masks,
            offset_height,
            offset_width,
            self.height,
            self.width,
        )
        segmentation_masks = tf.ensure_shape(
            segmentation_masks, shape=(None, self.height, self.width, None)
        )
        return tf.cast(segmentation_masks, dtype=self.compute_dtype)

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "height": self.height,
                "width": self.width,
                "padding_value": self.padding_value,
                "bounding_box_format": self.bounding_box_format,
                "bounding_box_min_area_ratio": self.bounding_box_min_area_ratio,  # noqa: E501
                "bounding_box_max_aspect_ratio": self.bounding_box_max_aspect_ratio,  # noqa: E501
                "seed": self.seed,
            }
        )
        return config



import keras
import tensorflow as tf

from keras_aug.core.factor_sampler.factor_sampler import FactorSampler
from keras_aug.core.keras_random_generator import KerasRandomGenerator


@keras.utils.register_keras_serializable(package="keras_aug")
class SignedConstantFactorSampler(FactorSampler):
    """SignedConstantFactorSampler samples the same factor for every call to
    ``__call__()`` and randomly inverts the sampled factors.

    This is useful in cases where a user wants to always ensure that an
    augmentation layer performs augmentations of the same strength.

    Args:
        value: the value to return from `__call__()`.
        rate (float, optional): The rate to invert factors. Defaults to ``0.5``.
        seed (int|float, optional): The random seed. Defaults to ``None``.

    References:
        - `KerasCV <https://github.com/keras-team/keras-cv>`_
    """

    def __init__(self, value, rate=0.5, seed=None):
        self.value = value
        self.rate = rate
        self.seed = seed
        self.rng = KerasRandomGenerator(seed=seed, rng_type="stateless")

    def __call__(self, shape=(), dtype="float32"):
        factors = tf.ones(shape=shape, dtype=dtype) * self.value
        negates = self.rng.random_uniform(
            shape=shape, minval=0, maxval=1, dtype=tf.float32
        )
        negates = tf.cast(tf.where(negates > self.rate, -1.0, 1.0), dtype=dtype)
        return factors * negates

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "value": self.value,
                "rate": self.rate,
                "seed": self.seed,
            }
        )
        return config



import tensorflow as tf
from tensorflow import keras

from keras_aug.datapoints import image as image_utils
from keras_aug.layers.base.vectorized_base_random_layer import (
    VectorizedBaseRandomLayer,
)
from keras_aug.utils import augmentation as augmentation_utils


@keras.utils.register_keras_serializable(package="keras_aug")
class RandomChannelShift(VectorizedBaseRandomLayer):
    """Randomly shift values for each channel of the input images.

    Args:
        value_range (Sequence[int|float]): The range of values the incoming
            images will have. This is typically either ``[0, 1]`` or
            ``[0, 255]`` depending on how your preprocessing pipeline is set up.
        factor (float|Sequence[float]|keras_aug.FactorSampler): The range of the
            channel shift factor. When represented as a single float,
            the factor will be picked between ``[0.0 - lower, 0.0 + upper]``.
            ``0.0`` gives the original image.
        channels (int, optional): The number of channels to shift. Defaults to
            ``3`` corresponds to RGB shift.
        seed (int|float, optional): The random seed. Defaults to
            ``None``.

    References:
        - `KerasCV <https://github.com/keras-team/keras-cv>`_
    """

    def __init__(self, value_range, factor, channels=3, seed=None, **kwargs):
        super().__init__(seed=seed, **kwargs)
        self.value_range = value_range
        self.factor = augmentation_utils.parse_factor(
            factor,
            min_value=-1.0,
            max_value=1.0,
            center_value=0.0,
        )
        self.channels = channels
        self.seed = seed

    def get_random_transformation_batch(self, batch_size, **kwargs):
        shifts = self.factor(
            shape=(batch_size, self.channels), dtype=self.compute_dtype
        )
        return shifts

    def augment_ragged_image(self, image, transformation, **kwargs):
        images = tf.expand_dims(image, axis=0)
        transformations = tf.expand_dims(transformation, axis=0)
        images = self.augment_images(
            images=images, transformations=transformations, **kwargs
        )
        return tf.squeeze(images, axis=0)

    def augment_images(self, images, transformations=None, **kwargs):
        images = image_utils.transform_value_range(
            images, self.value_range, (0, 1), dtype=self.compute_dtype
        )
        shifts = transformations[:, tf.newaxis, tf.newaxis, :]
        images = images + shifts
        images = tf.clip_by_value(images, 0.0, 1.0)
        images = image_utils.transform_value_range(
            images, (0, 1), self.value_range, dtype=self.compute_dtype
        )
        return images

    def augment_labels(self, labels, transformations, **kwargs):
        return labels

    def augment_bounding_boxes(self, bounding_boxes, transformations, **kwargs):
        return bounding_boxes

    def augment_segmentation_masks(
        self, segmentation_masks, transformations, **kwargs
    ):
        return segmentation_masks

    def augment_keypoints(self, keypoints, transformations, **kwargs):
        return keypoints

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "value_range": self.value_range,
                "factor": self.factor,
                "channels": self.channels,
                "seed": self.seed,
            }
        )
        return config



import tensorflow as tf
from absl.testing import parameterized
from tensorflow import keras

from keras_aug import layers
from keras_aug.layers.base.vectorized_base_random_layer import (
    VectorizedBaseRandomLayer,
)


class ZeroOut(VectorizedBaseRandomLayer):
    """Zero out all entries, for testing purposes."""

    def __init__(self):
        super(ZeroOut, self).__init__()

    def augment_images(self, image, transformation=None, **kwargs):
        return 0 * image

    def augment_labels(self, label, transformation=None, **kwargs):
        return 0 * label


class RandomApplyTest(tf.test.TestCase, parameterized.TestCase):
    rng = tf.random.Generator.from_seed(seed=1234)

    @parameterized.parameters([-0.5, 1.7])
    def test_raises_error_on_invalid_rate_parameter(self, invalid_rate):
        with self.assertRaises(ValueError):
            layers.RandomApply(rate=invalid_rate, layer=ZeroOut())

    def test_inputs_unchanged_with_zero_rate(self):
        dummy_inputs = self.rng.uniform(shape=(4, 16, 16, 3))
        layer = layers.RandomApply(rate=0.0, layer=ZeroOut())

        outputs = layer(dummy_inputs)

        self.assertAllClose(outputs, dummy_inputs)

    def test_all_inputs_changed_with_rate_equal_to_one(self):
        dummy_inputs = self.rng.uniform(shape=(4, 16, 16, 3))
        layer = layers.RandomApply(rate=1.0, layer=ZeroOut())

        outputs = layer(dummy_inputs)

        self.assertAllEqual(outputs, tf.zeros_like(dummy_inputs))

    def test_works_with_single_image(self):
        dummy_inputs = self.rng.uniform(shape=(16, 16, 3))
        layer = layers.RandomApply(rate=1.0, layer=ZeroOut())

        outputs = layer(dummy_inputs)

        self.assertAllEqual(outputs, tf.zeros_like(dummy_inputs))

    def test_can_modify_label(self):
        dummy_inputs = self.rng.uniform(shape=(4, 16, 16, 3))
        dummy_labels = tf.ones(shape=(4, 2))
        layer = layers.RandomApply(rate=1.0, layer=ZeroOut())

        outputs = layer({"images": dummy_inputs, "labels": dummy_labels})

        self.assertAllEqual(outputs["labels"], tf.zeros_like(dummy_labels))

    def test_works_with_native_keras_layers(self):
        dummy_inputs = self.rng.uniform(shape=(4, 16, 16, 3))
        zero_out = keras.layers.Lambda(lambda x: {"images": 0 * x["images"]})
        layer = layers.RandomApply(rate=1.0, layer=zero_out)

        outputs = layer(dummy_inputs)

        self.assertAllEqual(outputs, tf.zeros_like(dummy_inputs))

    def test_works_with_xla(self):
        dummy_inputs = self.rng.uniform(shape=(4, 16, 16, 3))
        # auto_vectorize=True will crash XLA
        layer = layers.RandomApply(rate=0.5, layer=ZeroOut())

        @tf.function(jit_compile=True)
        def apply(x):
            return layer(x)

        apply(dummy_inputs)



import tensorflow as tf
from tensorflow import keras

from keras_aug.layers.base.vectorized_base_random_layer import (
    VectorizedBaseRandomLayer,
)


@keras.utils.register_keras_serializable(package="keras_aug")
class ChannelShuffle(VectorizedBaseRandomLayer):
    """Shuffles channels of the input images.

    Args:
        groups (int, optional): The number of the groups to divide the input
            channels. Defaults to ``3``.
        seed (int|float, optional): The random seed. Defaults to
            ``None``.

    References:
        - `KerasCV <https://github.com/keras-team/keras-cv>`_
    """

    def __init__(self, groups=3, seed=None, **kwargs):
        super().__init__(seed=seed, **kwargs)
        self.groups = groups
        self.seed = seed

    def get_random_transformation_batch(self, batch_size, **kwargs):
        # get batched shuffled indices
        # for example: batch_size=2; self.group=5
        # indices = [
        #     [0, 2, 3, 4, 1],
        #     [4, 1, 0, 2, 3]
        # ]
        indices = self._random_generator.random_uniform(
            (batch_size, self.groups)
        )
        indices = tf.argsort(indices, axis=-1)
        return indices

    def augment_ragged_image(self, image, transformation, **kwargs):
        image = tf.expand_dims(image, axis=0)
        transformation = tf.expand_dims(transformation, axis=0)
        image = self.augment_images(
            images=image, transformations=transformation, **kwargs
        )
        return tf.squeeze(image, axis=0)

    def augment_images(self, images, transformations, **kwargs):
        batch_size = tf.shape(images)[0]
        height, width = tf.shape(images)[1], tf.shape(images)[2]
        num_channels = images.shape[3]
        indices = transformations

        # append batch indexes next to shuffled indices
        batch_indexs = tf.repeat(tf.range(batch_size), self.groups)
        batch_indexs = tf.reshape(batch_indexs, (batch_size, self.groups))
        indices = tf.stack([batch_indexs, indices], axis=-1)

        if not num_channels % self.groups == 0:
            raise ValueError(
                "The number of input channels should be "
                "divisible by the number of groups."
                f"Received: channels={num_channels}, groups={self.groups}"
            )
        images = tf.reshape(
            images, [batch_size, height, width, self.groups, -1]
        )
        images = tf.transpose(images, perm=[0, 3, 1, 2, 4])
        images = tf.gather_nd(images, indices=indices)
        images = tf.transpose(images, perm=[0, 2, 3, 4, 1])
        images = tf.reshape(images, [batch_size, height, width, num_channels])
        return images

    def augment_labels(self, labels, transformations, **kwargs):
        return labels

    def augment_bounding_boxes(self, bounding_boxes, transformations, **kwargs):
        return bounding_boxes

    def augment_segmentation_masks(
        self, segmentation_masks, transformations, **kwargs
    ):
        return segmentation_masks

    def augment_keypoints(self, keypoints, transformations, **kwargs):
        return keypoints

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "groups": self.groups,
                "seed": self.seed,
            }
        )
        return config



import tensorflow as tf
from tensorflow import keras

from keras_aug.layers.base.vectorized_base_random_layer import (
    VectorizedBaseRandomLayer,
)
from keras_aug.utils.augmentation import BOUNDING_BOXES
from keras_aug.utils.augmentation import IMAGES


@keras.utils.register_keras_serializable(package="keras_aug")
class RepeatedAugment(VectorizedBaseRandomLayer):
    """RepeatedAugment augments each image in a batch multiple times.

    This technique exists to emulate the behavior of stochastic gradient descent
    within the context of mini-batch gradient descent. When training large
    vision models, choosing a large batch size can introduce too much noise into
    aggregated gradients causing the overall batch's gradients to be less
    effective than gradients produced using smaller gradients.
    RepeatedAugment handles this by re-using the same image multiple times
    within a batch creating correlated samples.

    Notes:
        This layer increases your batch size by a factor of ``len(layers)``.

    Args:
        layers (list(keras_aug.layers.*)): The list of the layers to use to
            augment the inputs.
        shuffle (bool, optional): Whether to shuffle the results. Essential when
            using an asynchronous distribution strategy such as
            ParameterServerStrategy. Defaults to ``True``.
        seed (int|float, optional): The random seed. Defaults to ``None``.

    References:
        - `RepeatedAugment <https://arxiv.org/abs/1901.09335>`_
        - `DEIT <https://github.com/facebookresearch/deit>`_
        - `KerasCV <https://github.com/keras-team/keras-cv>`_
    """  # noqa: E501

    def __init__(self, layers, shuffle=True, seed=None, **kwargs):
        super().__init__(seed=seed, **kwargs)
        self.layers = layers
        self.shuffle = shuffle
        self.seed = seed

    def _batch_augment(self, inputs):
        layer_outputs = [layer(inputs) for layer in self.layers]

        results = {}
        for k in inputs.keys():
            if k == BOUNDING_BOXES:
                results[k] = {}
                results[k]["boxes"] = tf.concat(
                    [output[k]["boxes"] for output in layer_outputs], axis=0
                )
                results[k]["classes"] = tf.concat(
                    [output[k]["classes"] for output in layer_outputs], axis=0
                )
            else:
                results[k] = tf.concat(
                    [output[k] for output in layer_outputs], axis=0
                )
        if self.shuffle:
            shuffle_size = tf.shape(results[IMAGES])[0]
            results = self.shuffle_results(results, shuffle_size)
        return results

    def shuffle_results(self, results, shuffle_size):
        indices = tf.range(start=0, limit=shuffle_size, dtype=tf.int32)
        indices = tf.random.experimental.stateless_shuffle(
            indices, seed=self._random_generator.make_seed_for_stateless_op()
        )
        for k in results.keys():
            if k == BOUNDING_BOXES:
                results[k]["boxes"] = tf.gather(results[k]["boxes"], indices)
                results[k]["classes"] = tf.gather(
                    results[k]["classes"], indices
                )
            else:
                results[k] = tf.gather(results[k], indices)
        return results

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "layers": self.layers,
                "shuffle": self.shuffle,
                "seed": self.seed,
            }
        )
        return config



import tensorflow as tf
from tensorflow import keras

from keras_aug.datapoints import image as image_utils
from keras_aug.layers.base.vectorized_base_random_layer import (
    VectorizedBaseRandomLayer,
)


@keras.utils.register_keras_serializable(package="keras_aug")
class Equalize(VectorizedBaseRandomLayer):
    """Performs histogram equalization on a channel-wise basis.

    Args:
        value_range (Sequence[int|float]): The range of values the incoming
            images will have. This is typically either ``[0, 1]`` or
            ``[0, 255]`` depending on how your preprocessing pipeline is set up.
        bins (int, optional): The number of bins to use in histogram
            equalization. Should be in the range ``[0, 256]``. Defaults to
            ``256``.

    References:
        - `KerasCV <https://github.com/keras-team/keras-cv>`_
    """

    def __init__(self, value_range, bins=256, **kwargs):
        super().__init__(**kwargs)
        self.value_range = value_range
        self.bins = bins

    def augment_ragged_image(self, image, transformation, **kwargs):
        image = tf.expand_dims(image, axis=0)
        image = self.augment_images(
            images=image, transformations=transformation, **kwargs
        )
        return tf.squeeze(image, axis=0)

    def augment_images(self, images, transformations=None, **kwargs):
        original_shape = images.shape
        images = image_utils.transform_value_range(
            images, self.value_range, (0, 255), dtype=self.compute_dtype
        )
        images = tf.cast(images, dtype=tf.int32)
        images = tf.map_fn(
            self.equalize_single_image,
            images,
        )
        images = tf.transpose(images, (0, 2, 3, 1))
        images = tf.cast(images, dtype=self.compute_dtype)
        images = image_utils.transform_value_range(
            images, (0, 255), self.value_range, dtype=self.compute_dtype
        )
        images.set_shape(original_shape)
        return tf.cast(images, dtype=self.compute_dtype)

    def augment_labels(self, labels, transformations, **kwargs):
        return labels

    def augment_bounding_boxes(self, bounding_boxes, transformations, **kwargs):
        return bounding_boxes

    def augment_segmentation_masks(
        self, segmentation_masks, transformations, **kwargs
    ):
        return segmentation_masks

    def augment_keypoints(self, keypoints, transformations, **kwargs):
        return keypoints

    def equalize_single_image(self, image):
        return tf.map_fn(
            lambda channel_index: self.equalize_single_channel(
                image, channel_index
            ),
            tf.range(tf.shape(image)[-1]),
        )

    def equalize_single_channel(self, image, channel_index):
        image = image[..., channel_index]
        # Compute the histogram of the image channel.
        histogram = tf.histogram_fixed_width(image, [0, 255], nbins=self.bins)
        # For the purposes of computing the step, filter out the non-zeros.
        # Zeroes are replaced by a big number while calculating min to keep
        # shape constant across input sizes for compatibility with
        # vectorized_map
        big_number = 1410065408
        histogram_without_zeroes = tf.where(
            tf.equal(histogram, 0),
            big_number,
            histogram,
        )
        step = (
            tf.reduce_sum(histogram) - tf.reduce_min(histogram_without_zeroes)
        ) // (self.bins - 1)

        def build_mapping(histogram, step):
            # Compute the cumulative sum, shifting by step // 2
            # and then normalization by step.
            lookup_table = (tf.cumsum(histogram) + (step // 2)) // step
            # Shift lookup_table, prepending with 0.
            lookup_table = tf.concat([[0], lookup_table[:-1]], 0)
            # Clip the counts to be in range. This is done
            # in the C code for image.point.
            return tf.clip_by_value(lookup_table, 0, 255)

        # If step is zero, return the original image. Otherwise, build
        # lookup table from the full histogram and step and then index from it.
        image = tf.cond(
            tf.equal(step, 0),
            lambda: image,
            lambda: tf.gather(build_mapping(histogram, step), image),
        )
        return image

    def get_config(self):
        config = super().get_config()
        config.update({"value_range": self.value_range, "bins": self.bins})
        return config



import numpy as np
import tensorflow as tf

from keras_aug import layers


class RandomPosterizeTest(tf.test.TestCase):
    rng = tf.random.Generator.from_non_deterministic_state()

    def test_raises_error_on_invalid_bits_parameter(self):
        invalid_values = [-1, 0, 9, 24]
        for value in invalid_values:
            with self.assertRaises(ValueError):
                layers.RandomPosterize(value_range=(0, 1), factor=value)

    def test_single_image(self):
        bits = self._get_random_bits()
        dummy_input = self.rng.uniform(shape=(4, 4, 3), maxval=256)
        expected_output = self._calc_expected_output(dummy_input, bits=bits)

        layer = layers.RandomPosterize(
            value_range=(0, 255), factor=(bits, bits)
        )
        output = layer(dummy_input)

        self.assertAllEqual(output, expected_output)

    def _get_random_bits(self):
        return int(
            self.rng.uniform(shape=(), minval=1, maxval=9, dtype=tf.int32)
        )

    def test_single_image_rescaled(self):
        bits = self._get_random_bits()
        dummy_input = self.rng.uniform(shape=(4, 4, 3), maxval=1.0)
        expected_output = (
            self._calc_expected_output(dummy_input * 255, bits=bits) / 255
        )

        layer = layers.RandomPosterize(value_range=[0, 1], factor=(bits, bits))
        output = layer(dummy_input)

        self.assertAllClose(output, expected_output)

    def test_batched_input(self):
        bits = self._get_random_bits()
        dummy_input = self.rng.uniform(shape=(2, 4, 4, 3), maxval=256)

        expected_output = []
        for image in dummy_input:
            expected_output.append(self._calc_expected_output(image, bits=bits))
        expected_output = tf.stack(expected_output)

        layer = layers.RandomPosterize(
            value_range=(0, 255), factor=(bits, bits)
        )
        output = layer(dummy_input)

        self.assertAllEqual(output, expected_output)

    @staticmethod
    def _calc_expected_output(image, bits):
        """layers.RandomPosterize in numpy, based on Albumentations:
        The algorithm is basically:
        1. create a lookup table of all possible input pixel values to pixel
            values after RandomPosterize
        2. map each pixel in the input to created lookup table.

        Source:
            https://github.com/albumentations-team/albumentations/blob/89a675cbfb2b76f6be90e7049cd5211cb08169a5/albumentations/augmentations/functional.py#L407
        """  # noqa: E501
        dtype = image.dtype
        image = tf.cast(image, tf.uint8)

        lookup_table = np.arange(0, 256, dtype=np.uint8)
        mask = ~np.uint8(2 ** (8 - bits) - 1)
        lookup_table &= mask

        return tf.cast(lookup_table[image], dtype)



import tensorflow as tf

from keras_aug import layers
from keras_aug.utils.augmentation import BOUNDING_BOXES
from keras_aug.utils.augmentation import IMAGES


class SanitizeBoundingBoxTest(tf.test.TestCase):
    def test_sanitize_small_bounding_boxes(self):
        images = tf.random.uniform((2, 512, 512, 3))
        bounding_boxes = {
            "boxes": tf.ragged.constant(
                [
                    [[10, 10, 20, 20], [100, 100, 150, 150]],
                    [[200, 200, 400, 400]],
                ],
                dtype=tf.float32,
            ),
            "classes": tf.ragged.constant([[0, 1], [2]], dtype=tf.float32),
        }
        expected_classes = tf.ragged.constant([[1], [2]], dtype=tf.float32)
        layer = layers.SanitizeBoundingBox(
            min_size=35, bounding_box_format="xyxy"
        )

        output = layer({IMAGES: images, BOUNDING_BOXES: bounding_boxes})

        self.assertAllEqual(output[BOUNDING_BOXES]["classes"], expected_classes)



import tensorflow as tf
from tensorflow import keras

from keras_aug.layers.base.vectorized_base_random_layer import (
    VectorizedBaseRandomLayer,
)


@keras.utils.register_keras_serializable(package="keras_aug")
class Rescale(VectorizedBaseRandomLayer):
    """Rescales the inputs to a new range.

    Rescale rescales every value of the inputs (often the images) by the
    equation: ``y = x * scale + offset``.

    Args:
        scale (int|float): The scale to apply to the inputs.
        offset (int|float, optional): The offset to apply to the inputs.
            Defaults to ``0.0``

    References:
        - `KerasCV <https://github.com/keras-team/keras-cv>`_
    """

    def __init__(self, scale, offset=0.0, **kwargs):
        super().__init__(**kwargs)
        self.scale = tf.cast(scale, self.compute_dtype)
        self.offset = tf.cast(offset, self.compute_dtype)

    def augment_ragged_image(self, image, transformation, **kwargs):
        images = tf.expand_dims(image, axis=0)
        images = self.augment_images(
            images=images, transformations=transformation, **kwargs
        )
        return tf.squeeze(images, axis=0)

    def augment_images(self, images, transformations, **kwargs):
        images = tf.cast(images, dtype=self.compute_dtype)
        return images * self.scale + self.offset

    def augment_labels(self, labels, transformations, **kwargs):
        return labels

    def augment_bounding_boxes(self, bounding_boxes, transformations, **kwargs):
        return bounding_boxes

    def augment_segmentation_masks(
        self, segmentation_masks, transformations, **kwargs
    ):
        return segmentation_masks

    def augment_keypoints(self, keypoints, transformations, **kwargs):
        return keypoints

    def get_config(self):
        config = super().get_config()
        config.update({"scale": self.scale, "offset": self.offset})
        return config
